#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import numpy as np
from astropy.io import fits


def save_ndarray_to_fits(array=None, file_name=None,
                         main_header=None,
                         cast_to_float=True,
                         crpix1=None, crval1=None, cdelt1=None,
                         overwrite=True):
    """Save numpy array(s) into a FITS file with the provided filename.

    Parameters
    ----------
    array : numpy array or list of numpy arrays
        Array(s) to be exported as the FITS file. If the input is
        a list, a multi-extension FITS file is generated assuming
        that the list contains a list of arrays.
    file_name : string
        File name for the output FITS file.
    main_header : astropy FITS header
        Header to be introduced in the primary HDU.
    cast_to_float : bool or list of booleans
        If True, the array(s) data are save as float. If a list of
        arrays has been provided, this parameter must be either a list
        (with the same length) of booleans or None.
    crpix1 : float, list of floats or None
        If not None, this value is used for the keyword CRPIX1.
        If a list of arrays has been provided, this paramater must be
        either a list (with the same length) of floats or None.
    crval1 : float, list of floats or None
        If not None, this value is used for the keyword CRVAL1.
        If a list of arrays has been provided, this paramater must be
        either a list (with the same length) of floats or None.
    cdelt1 : float, list of floats or None
        If not None, this value is used for the keyword CDELT1.
        If a list of arrays has been provided, this paramater must be
        either a list (with the same length) of floats or None.
    overwrite : bool
        If True, the file is overwritten (in the case it already
        exists).

    """

    # protections
    if file_name is None:
        raise ValueError("File_name is not defined in save_ndarray_to_fits")

    if type(array) is list:
        list_of_arrays = array
        narrays = len(list_of_arrays)
        # cast_to_float must be a list of bools
        if type(cast_to_float) is not list:
            raise ValueError("Expected list of cast_to_float not found!")
        else:
            if len(cast_to_float) != narrays:
                raise ValueError("Unexpected length of cast_to_float")
        list_cast_to_float = cast_to_float
        # check that the additional associated lists have been provided
        # and that they have the expected length (or they are None)
        for ldum, cdum in zip([crpix1, crval1, cdelt1],
                              ['crpix1', 'crval1', 'cdelt1']):
            if ldum is not None:
                if type(ldum) is not list:
                    raise ValueError("Expected list of " + cdum +
                                     " not found!")
                else:
                    if len(ldum) != narrays:
                        raise ValueError("Unexpected length of " + cdum)
        if crpix1 is None:
            list_crpix1 = [None] * narrays
        else:
            list_crpix1 = crpix1
        if crval1 is None:
            list_crval1 = [None] * narrays
        else:
            list_crval1 = crval1
        if cdelt1 is None:
            list_cdelt1 = [None] * narrays
        else:
            list_cdelt1 = cdelt1
    else:
        list_of_arrays = [array]
        list_cast_to_float = [cast_to_float]
        list_crpix1 = [crpix1]
        list_crval1 = [crval1]
        list_cdelt1 = [cdelt1]

    hdulist = fits.HDUList()

    for ihdu, tmp_array in enumerate(list_of_arrays):
        if type(tmp_array) is not np.ndarray:
            raise ValueError("Array#" + str(ihdu) + "=" + str(tmp_array) +
                             " must be a numpy.ndarray")
        if ihdu == 0:
            if list_cast_to_float[ihdu]:
                hdu = fits.PrimaryHDU(data=tmp_array.astype(np.float32),
                                      header=main_header)
            else:
                hdu = fits.PrimaryHDU(data=tmp_array, header=main_header)
        else:
            if list_cast_to_float[ihdu]:
                hdu = fits.ImageHDU(data=tmp_array.astype(np.float32))
            else:
                hdu = fits.ImageHDU(data=tmp_array)

        # set additional FITS keywords if requested
        tmp_crpix1 = list_crpix1[ihdu]
        if tmp_crpix1 is not None:
            hdu.header.set('CRPIX1', tmp_crpix1, 'Reference pixel')
        tmp_crval1 = list_crval1[ihdu]
        if tmp_crval1 is not None:
            hdu.header.set('CRVAL1', tmp_crval1,
                           'Reference wavelength corresponding to CRPIX1')
        tmp_cdelt1 = list_cdelt1[ihdu]
        if tmp_cdelt1 is not None:
            hdu.header.set('CDELT1', tmp_cdelt1,
                           'Linear dispersion (angstrom/pixel)')

        # add HDU to HDUList
        hdulist.append(hdu)

    # write output file
    hdulist.writeto(file_name, overwrite=overwrite)



#
# Copyright 2015-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Bar detection recipe for EMIR"""

from __future__ import division

import logging

from numina.array.utils import coor_to_pix_1d
from numina.core import Requirement, Result, Parameter
import numina.exceptions
import numina.types.array as tarray
from numina.processing.combine import basic_processing_with_combination
from scipy.ndimage.filters import median_filter
from skimage.feature import canny

import emirdrp.datamodel as datamodel
from emirdrp.core import EMIR_PIXSCALE
from emirdrp.core.recipe import EmirRecipe
import emirdrp.requirements as reqs
import emirdrp.products as prods
from emirdrp.processing.bardetect import find_position
from emirdrp.processing.bardetect import locate_bar_l, locate_bar_r
from .common import normalize_raw


class BarDetectionRecipe(EmirRecipe):

    # Recipe Requirements
    #
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement()

    bars_nominal_positions = Requirement(prods.CoordinateList2DType,
                                         'Nominal positions of the bars'
                                         )
    median_filter_size = Parameter(5, 'Size of the median box')
    canny_sigma = Parameter(3.0, 'Sigma for the canny algorithm')
    canny_high_threshold = Parameter(0.04, 'High threshold for the canny algorithm')
    canny_low_threshold = Parameter(0.01, 'High threshold for the canny algorithm')

    # Recipe Results
    frame = Result(prods.ProcessedImage)
    positions = Result(tarray.ArrayType)
    DTU = Result(tarray.ArrayType)
    ROTANG = Result(float)
    csupos = Result(tarray.ArrayType)
    csusens = Result(tarray.ArrayType)
    param_median_filter_size = Result(float)
    param_canny_high_threshold = Result(float)
    param_canny_low_threshold = Result(float)

    def run(self, rinput):

        logger = logging.getLogger('numina.recipes.emir')

        logger.info('starting processing for bars detection')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(rinput, flow=flow)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        try:
            rotang = hdr['ROTANG']
            mecs_hdr = datamodel.get_mecs_header(hdulist)
            dtub, dtur = datamodel.get_dtur_from_header(mecs_hdr)
            csupos = datamodel.get_csup_from_header(mecs_hdr)
            csusens = datamodel.get_cs_from_header(mecs_hdr)
        except KeyError as error:
            logger.error(error)
            raise numina.exceptions.RecipeError(error)

        logger.debug('finding bars')

        arr = hdulist[0].data

        # Median filter
        logger.debug('median filtering')
        mfilter_size = rinput.median_filter_size

        arr_median = median_filter(arr, size=mfilter_size)

        # Image is mapped between 0 and 1
        # for the full range [0: 2**16]
        logger.debug('image scaling to 0-1')
        arr_grey = normalize_raw(arr_median)

        # Find borders
        logger.debug('find borders')
        canny_sigma = rinput.canny_sigma
        # These threshols corespond roughly with
        # value x (2**16 - 1)
        high_threshold = rinput.canny_high_threshold
        low_threshold = rinput.canny_low_threshold

        edges = canny(arr_grey, sigma=canny_sigma,
                      high_threshold=high_threshold,
                      low_threshold=low_threshold)

        # Number or rows used
        # These other parameters cab be tuned also
        total = 5
        maxdist = 1.0
        bstart = 100
        bend = 1900

        positions = []
        nt = total // 2

        xfac = dtur[0] / EMIR_PIXSCALE
        yfac = -dtur[1] / EMIR_PIXSCALE

        vec = [yfac, xfac]
        logger.debug('DTU shift is %s', vec)

        # Based on the 'edges image'
        # and the table of approx positions of the slits
        barstab = rinput.bars_nominal_positions

        # Currently, we only use fields 0 and 2
        # of the nominal positions file

        for coords in barstab:
            lbarid = int(coords[0])
            rbarid = lbarid + 55
            ref_y_coor = coords[2] + vec[1]
            prow = coor_to_pix_1d(ref_y_coor) - 1
            fits_row = prow + 1 # FITS pixel index

            logger.debug('looking for bars with ids %d - %d', lbarid, rbarid)
            logger.debug('reference y position is Y %7.2f', ref_y_coor)
            # Find the position of each bar

            bpos = find_position(edges, prow, bstart, bend, total)

            nbars_found = len(bpos)

            # If no bar is found, append and empty token
            if nbars_found == 0:
                logger.debug('bars %d, %d not found at row %d', lbarid, rbarid, fits_row)
                thisres1 = (lbarid, fits_row, 0, 0, 1)
                thisres2 = (rbarid, fits_row, 0, 0, 1)

            elif nbars_found == 2:

                # Order values by increasing X
                centl, centr = sorted(bpos, key=lambda cen: cen[0])
                c1 = centl[0]
                c2 = centr[0]

                logger.debug('bars found  at row %d between %7.2f - %7.2f', fits_row, c1, c2)
                # Compute FWHM of the collapsed profile

                cslit = arr_grey[prow-nt:prow+nt+1,:]
                pslit = cslit.mean(axis=0)

                # Add 1 to return FITS coordinates
                epos, epos_f, error = locate_bar_l(pslit, c1)
                thisres1 = lbarid, fits_row, epos + 1, epos_f + 1, error


                epos, epos_f, error = locate_bar_r(pslit, c2)
                thisres2 = rbarid, fits_row, epos + 1, epos_f + 1, error

            elif nbars_found == 1:
                logger.warning('only 1 edge found  at row %d, not yet implemented', fits_row)
                thisres1 = (lbarid, fits_row, 0, 0, 1)
                thisres2 = (rbarid, fits_row, 0, 0, 1)

            else:
                logger.warning('3 or more edges found  at row %d, not yet implemented', fits_row)
                thisres1 = (lbarid, fits_row, 0, 0, 1)
                thisres2 = (rbarid, fits_row, 0, 0, 1)

            positions.append(thisres1)
            positions.append(thisres2)

        logger.debug('end finding bars')
        result = self.create_result(frame=hdulist,
                                    positions=positions,
                                    DTU=dtub,
                                    ROTANG=rotang,
                                    csupos=csupos,
                                    csusens=csusens,
                                    param_median_filter_size=rinput.median_filter_size,
                                    param_canny_high_threshold=rinput.canny_high_threshold,
                                    param_canny_low_threshold=rinput.canny_low_threshold
                                    )
        return result



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Data products produced by the EMIR pipeline."""

import uuid
import copy

import six
import numpy

from numina.ext.gtc import DF
from numina.core import DataFrameType, DataProductType
import numina.types.array as arrtype
from numina.types.linescatalog import LinesCatalog
import numina.exceptions
import numina.types.product as prodtypes
import numina.types.structured
import numina.types.obsresult as obtypes

import emirdrp.datamodel

base_schema_description = {
    'keywords': {
        'INSTRUME': {'mandatory': True},
        'READMODE': {'mandatory': True,
                     'value': ['SIMPLE', 'BIAS', 'SINGLE',
                               'CDS', 'FOWLER', 'RAMP']
                     },
        'EXPTIME': {'value': float},
        'NUMINAID': {'value': int}
        }
    }

gtc_proc_schema_description = {
    'keywords': {
        'NUMINAID': {'mandatory': True, 'value': int}
        }
    }

emir_schema_description = {
    'keywords': {
        'INSTRUME': {'mandatory': True, 'value': 'EMIR'},
        'READMODE': {'mandatory': True,
                     'value': ['SIMPLE', 'BIAS', 'SINGLE',
                               'CDS', 'FOWLER', 'RAMP']
                     },
        'BUNIT': {'value': ['ADU', 'ADU/s']},
        'IMGTYPE': {'mandatory': True, 'type': 'string'},
        }
    }


class EmirFrame(DataFrameType):

    def __init__(self, *args, **kwargs):
        super(EmirFrame, self).__init__(datamodel=emirdrp.datamodel.EmirDataModel)


class ProcessedFrame(EmirFrame):
    """A processed frame"""
    pass


class ProcessedImage(ProcessedFrame):
    """A processed image"""
    pass


class ProcessedMOS(ProcessedFrame):
    """A processed image with slit spectra"""
    pass


class ProcessedSpectrum(ProcessedFrame):
    """A 1d spectrum"""
    pass


class ProcessedImageProduct(prodtypes.DataProductMixin, ProcessedImage):
    pass


class ProcessedMOSProduct(prodtypes.DataProductMixin, ProcessedMOS):
    pass


class ProcessedSpectrumProduct(prodtypes.DataProductMixin, ProcessedSpectrum):
    pass


class MasterBadPixelMask(ProcessedImageProduct):
    pass


class MasterBias(ProcessedImageProduct):
    """Master bias product
    """
    pass


class MasterDark(ProcessedImageProduct):
    """Master dark product
    """
    pass


class SkyLinesCatalog(LinesCatalog):
    """Sky Lines Catalog
    """
    pass


class MasterIntensityFlat(ProcessedImageProduct):
    __tags__ = ['filter']

class MasterSpectralFlat(ProcessedImageProduct):
    __tags__ = ['grism', 'filter']


# FIXME: This is not really a calibration
class MasterSky(ProcessedImageProduct):
    __tags__ = ['filter']


# FIXME: This is not really a calibration
class SkySpectrum(ProcessedImageProduct):
    __tags__ = ['grism', 'filter']


class Spectra(DataFrameType):
    pass


class DataCube(DataFrameType):
    pass


class TelescopeFocus(DataProductType):
    def __init__(self, default=None):
        super(TelescopeFocus, self).__init__(ptype=float)


class DTUFocus(DataProductType):
    def __init__(self, default=None):
        super(DTUFocus, self).__init__(ptype=float)


class DTU_XY_Calibration(DataFrameType):
    pass


class DTU_Z_Calibration(DataFrameType):
    pass


class DTUFlexureCalibration(DataFrameType):
    pass


# FIXME:
class SlitTransmissionCalibration(DataFrameType):
    pass


# FIXME:
class WavelengthCalibration(DataFrameType):
    pass


class CSU2DetectorCalibration(DataFrameType):
    pass


class PointingOriginCalibration(DataFrameType):
    pass


class SpectroPhotometricCalibration(DataFrameType):
    pass


class PhotometricCalibration(DataFrameType):
    pass


class MasterGainMap(DataProductType):
    def __init__(self, mean, var, frame):
        self.mean = mean
        self.var = var
        self.frame = frame

    def __getstate__(self):
        gmean = list(map(float, self.mean.flat))
        gvar = list(map(float, self.var.flat))
        return {'frame': self.frame, 'mean': gmean, 'var': gvar}


class MasterRONMap(DataProductType):
    def __init__(self, mean, var):
        self.mean = mean
        self.var = var

    def __getstate__(self):
        gmean = map(float, self.mean.flat)
        gvar = map(float, self.var.flat)
        return {'mean': gmean, 'var': gvar}


class TelescopeOffset(DataProductType):
    def __init__(self, default=None):
        super(TelescopeOffset, self).__init__(ptype=float)


class CoordinateListNType(arrtype.ArrayNType):
    def __init__(self, dimensions, default=None):
        super(CoordinateListNType,
              self).__init__(dimensions,
                             default=default)

    def validate(self, obj):
        ndims = len(obj.shape)
        if ndims != 2:
            raise numina.exceptions.ValidationError(
                '%r is not a valid %r' %
                (obj, self.__class__.__name__)
            )
        if obj.shape[1] != self.N:
            raise numina.exceptions.ValidationError(
                '%r is not a valid %r' %
                (obj, self.__class__.__name__)
            )


class CoordinateList1DType(CoordinateListNType):
    def __init__(self, default=None):
        super(CoordinateList1DType, self).__init__(1, default=default)


class CoordinateList2DType(CoordinateListNType):
    def __init__(self, default=None):
        super(CoordinateList2DType, self).__init__(2, default=default)
        self.add_dialect_info('gtc', DF.TYPE_DOUBLE_ARRAY2D)


class NominalPositions(prodtypes.DataProductMixin, CoordinateList2DType):
    pass


class MSMPositions(DataProductType):
    def __init__(self):
        super(MSMPositions, self).__init__(ptype=list)


class SourcesCatalog(DataProductType):
    def __init__(self):
        super(SourcesCatalog, self).__init__(ptype=list)


class CentroidsTableType(DataProductType):
    """Table with information about focus centroids."""
    def __init__(self):
        super(CentroidsTableType, self).__init__(ptype=numpy.ndarray)


class ChannelLevelStatistics(DataProductType):
    """A list of exposure time, mean, std dev and median per channel"""
    def __init__(self, exposure, statistics):
        self.exposure = exposure
        self.statistics = statistics

    def __numina_dump__(self, obj, where):
        fname = 'statistics.txt'

        header = ("Channel Level Statistics\n"
                  "comment 2\n"
                  "pixels start in 1\n"
                  "pixels end in 2048\n"
                  "exposure={exposure}\n"
                  "xbegin xend ybegin yend mean median var\n"
                  )

        inter = header.format(exposure=obj.exposure)
        numpy.savetxt(fname, obj.statistics, header=inter)
        return fname



class ChannelLevelStatisticsType(DataProductType):
    """A list of exposure time, mean, std dev and median per channel"""
    def __init__(self):
        super(ChannelLevelStatisticsType,
              self).__init__(ptype=ChannelLevelStatistics)


class LinesCatalog(DataProductType):
    def __init__(self):
        super(LinesCatalog, self).__init__(ptype=numpy.ndarray)


class RefinedBoundaryModelParam(numina.types.structured.BaseStructuredCalibration):
    """Refined parameters of MOS model
    """
    def __init__(self, instrument='unknown'):
        super(RefinedBoundaryModelParam, self).__init__(instrument)
        self.tags = {
            'grism': "unknown",
            'filter': "unknown"
        }
        self.contents = []

    def __getstate__(self):
        state = super(RefinedBoundaryModelParam, self).__getstate__()
        if six.PY2:
            state['contents'] = copy.copy(self.contents)
        else:
            state['contents'] = self.contents.copy()
        return state

    def __setstate__(self, state):
        super(RefinedBoundaryModelParam, self).__setstate__(state)
        if six.PY2:
            self.contents = copy.copy(state['contents'])
        else:
            self.contents = state['contents'].copy()

    def tag_names(self):
        return ['grism', 'filter']


class RectWaveCoeff(numina.types.structured.BaseStructuredCalibration):
    """Rectification and Wavelength Calibration Coefficients
    """
    def __init__(self, instrument='unknown'):
        super(RectWaveCoeff, self).__init__(instrument)
        self.tags = {
            'grism': "unknown",
            'filter': "unknown"
        }
        self.global_integer_offset_x_pix = 0
        self.global_integer_offset_y_pix = 0
        self.total_slitlets = 0
        self.missing_slitlets = []
        self.contents = []

    def __getstate__(self):
        state = super(RectWaveCoeff, self).__getstate__()

        keys = ['global_integer_offset_x_pix', 'global_integer_offset_y_pix',
                'total_slitlets', 'missing_slitlets']
        for key in keys:
            state[key] = self.__dict__[key]

        if six.PY2:
            state['contents'] = copy.copy(self.contents)
        else:
            state['contents'] = self.contents.copy()
        return state

    def __setstate__(self, state):
        super(RectWaveCoeff, self).__setstate__(state)

        keys = ['global_integer_offset_x_pix', 'global_integer_offset_y_pix',
                'total_slitlets', 'missing_slitlets']
        for key in keys:
            self.__dict__[key] = state[key]
        if six.PY2:
            self.contents = copy.copy(state['contents'])
        else:
            self.contents = state['contents'].copy()

    def tag_names(self):
        return ['grism', 'filter']


class MasterRectWave(numina.types.structured.BaseStructuredCalibration):
    """Rectification and Wavelength Calibration Library Product
    """
    def __init__(self, instrument='unknown'):
        import numina.core.tagexpr as tagexpr

        datamodel = emirdrp.datamodel.EmirDataModel()
        super(MasterRectWave, self).__init__(instrument, datamodel=datamodel)
        self.tags = {
            'grism': "unknown",
            'filter': "unknown"
        }

        my_tag_table = self.datamodel.query_attrs
        objtags = [my_tag_table[t] for t in self.tag_names()]
        self.query_expr = tagexpr.query_expr_from_attr(objtags)
        self.names_t = self.query_expr.tags()
        self.names_f = self.query_expr.fields()
        self.query_opts = []

        self.total_slitlets = 0
        self.missing_slitlets = []
        self.contents = []

    def __getstate__(self):
        state = super(MasterRectWave, self).__getstate__()

        keys = ['total_slitlets', 'missing_slitlets']
        for key in keys:
            state[key] = self.__dict__[key]

        if six.PY2:
            # list has no .copy method in PY2
            state['contents'] = copy.copy(self.contents)
        else:
            state['contents'] = self.contents.copy()
        return state

    def __setstate__(self, state):
        from numina.types.typedialect import dialect_info
        import numina.core.tagexpr as tagexpr

        super(MasterRectWave, self).__setstate__(state)

        # Additional values
        # These values are not set with setstate
        self.datamodel = emirdrp.datamodel.EmirDataModel()
        self.internal_default = None
        self.internal_type = self.__class__
        self.internal_dialect = dialect_info(self)
        #
        my_tag_table = self.datamodel.query_attrs

        objtags = [my_tag_table[t] for t in self.tag_names()]
        self.query_expr = tagexpr.query_expr_from_attr(objtags)
        self.names_t = self.query_expr.tags()
        self.names_f = self.query_expr.fields()
        self.query_opts = []

        keys = ['total_slitlets', 'missing_slitlets']
        for key in keys:
            self.__dict__[key] = state[key]

        if six.PY2:
            # list has no .copy method in PY2
            self.contents = copy.copy(state['contents'])
        else:
            self.contents = state['contents'].copy()

    def tag_names(self):
        return ['grism', 'filter']

try:
    # FIXME: put this where it makes sense
    from gtc.DSL.DGCSTypes.IDL_Adapters import toIDL_, toIDL_struct, toElementType

    toIDL_.register(MasterRectWave, toIDL_struct)

    @toElementType.register(MasterRectWave)
    def _(value):
        return DF.TYPE_STRUCT
except ImportError:
    pass


if __name__ == '__main__':
    m = MasterRectWave().query_expr
    print(m.tags())
    m = MasterIntensityFlat().query_expr
    print(m.tags())



#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Definition of Slitlet2dArc class"""

from __future__ import division, print_function

from matplotlib.patches import Rectangle
import numpy as np
from scipy import ndimage
from skimage import restoration

from numina.array.ccd_line import ArcLine
from numina.array.ccd_line import intersection_spectrail_arcline
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.polfit_residuals import \
    polfit_residuals_with_sigma_rejection
from numina.array.display.ximplotxy import ximplotxy
from numina.array.display.ximshow import ximshow
from numina.array.distortion import compute_distortion
from numina.array.distortion import fmap
from numina.array.distortion import rectify2d
from numina.array.wavecalib.peaks_spectrum import find_peaks_spectrum
from numina.array.wavecalib.peaks_spectrum import refine_peaks_spectrum

from emirdrp.tools.fit_boundaries import expected_distorted_boundaries
from emirdrp.tools.fit_boundaries import expected_distorted_frontiers

from .rescale_array_z1z2 import rescale_array_to_z1z2
from .rescale_array_z1z2 import rescale_array_from_z1z2

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from emirdrp.core import EMIR_NPIXPERSLIT_RECTIFIED


def expected_y0_lower_frontier(islitlet):
    """Expected ordinate of lower frontier in rectified image.

    Parameters
    ----------
    islitlet : int
        Slitlet number.

    Returns
    ----------
    y0_lower : float
        Ordinate value.
    """

    y0_lower =  0.5 + float((islitlet - 1) * EMIR_NPIXPERSLIT_RECTIFIED)
    return y0_lower


def expected_y0_upper_frontier(islitlet):
    """Expected ordinate of upper frontier in rectified image.

    Parameters
    ----------
    islitlet : int
        Slitlet number.

    Returns
    ----------
    y0_upper : float
        Ordinate value.
    """

    y0_upper =  float((islitlet) * EMIR_NPIXPERSLIT_RECTIFIED) + 0.5
    return y0_upper


class Slitlet2dArc(object):
    """Slitlet2dArc class definition.

    Slitlet2dArc: 2D slitlet class for Long-Slit Arc image

    It is important to distinguish between boundaries (the slitlet
    region when useful information is available) and frontiers (which
    define the real separation between consecutive slitlets when no gap
    between them is considered).

    Parameters
    ----------
    islitlet : int
        Slitlet number.
    csu_conf : CsuConfiguration object
        Instance of CsuConfiguration.
    ymargin_bb : int
        Extra number of pixels above and below the enclosing rectangle
        defined by the slitlet frontiers.
    params : :class:`~lmfit.parameter.Parameters` or None
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str or None
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

   Attributes
    ----------
    islitlet : int
        Slitlet number.
    csu_bar_left : float
        Location (mm) of the left bar for each slitlet.
    csu_bar_right : list of floats
        Location (mm) of the right bar for each slitlet, using the
        same origin employed for csu_bar_left (which is not the
        value stored in the FITS keywords.
    csu_bar_slit_center : list of floats
        Middle point (mm) in between the two bars defining a slitlet.
    csu_bar_slit_width : list of floats
        Slitlet width (mm), computed as the distance between the two
        bars defining the slitlet.
    ymargin_bb : int
        Extra number of pixels above and below the enclosing rectangle
        defined by the slitlet frontiers.
    bb_nc1_orig : int
        Minimum X coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    bb_nc2_orig : int
        Maximum X coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    bb_ns1_orig : int
        Minimum Y coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    bb_ns2_orig : int
        Maximum Y coordinate of the enclosing bounding box (in pixel
        units) in the original image.
    x0_reference : float
        X coordinate where the rectified y0_reference_middle is computed
        as the Y coordinate of the middle spectrum trail. The same value
        is used for all the available spectrum trails.
    list_spectrails: list of SpectrumTrail instances
        List of spectrum trails defined.
    list_frontiers: list of SpectrumTrail instances
        List of spectrum trails defining the slitlet frontiers.
    y0_reference_lower: float
        Y coordinate corresponding to the lower spectrum trail computed
        at x0_reference. This value is employed as the Y coordinate of
        the lower spectrum trail of the rectified slitlet.
    y0_reference_middle: float
        Y coordinate corresponding to the middle spectrum trail computed
        at x0_reference. This value is employed as the Y coordinate of
        the middle spectrum trail of the rectified slitlet.
    y0_reference_upper: float
        Y coordinate corresponding to the upper spectrum trail computed
        at x0_reference. This value is employed as the Y coordinate of
        the upper spectrum trail of the rectified slitlet.
    y0_frontier_lower: float
        Y coordinate corresponding to the lower frontier computed at
        x0_reference.
    y0_frontier_upper: float
        Y coordinate corresponding to the upper frontier computed at
        x0_reference.
    y0_reference_lower_expected: float
        Expected Y coordinate corresponding to the lower boundary
        computed at x0_reference in the rectified image.
    y0_reference_middle_expected: float
        Expected Y coordinate corresponding to the middle boundary
        computed at x0_reference in the rectified image.
    y0_reference_upper_expected: float
        Expected Y coordinate corresponding to the upper boundary
        computed at x0_reference in the rectified image.
    y0_frontier_lower_expected: float
        Expected Y coordinate corresponding to the lower frontier
        computed at x0_reference in the rectified image.
    y0_frontier_upper_expected: float
        Expected Y coordinate corresponding to the upper frontier
        computed at x0_reference in the rectified image.
    corr_yrect_a : float
        Intercept of the relation y_expected = a + b * y_measured
        that transforms the measured ordinate into the expected ordinate
        of the rectified image.
    corr_yrect_b : float
        Slope of the relation y_expected = a + b * y_measured
        that transforms the measured ordinate into the expected ordinate
        of the rectified image.
    min_row_rectified : int
        Minimum useful row (starting from zero) of the rectified slitlet.
    max_row_rectified : int
        Maximum useful row (starting from zero) of the rectified slitlet.
    x_inter_orig : 1d numpy array, float
        X coordinates of the intersection points of arc lines with
        spectrum trails in the original image.
    y_inter_orig : 1d numpy array, float
        Y coordinates of the intersection points of arc lines with
        spectrum trails in the original image.
    x_inter_rect : 1d numpy array, float
        X coordinates of the intersection points of arc lines with
        spectrum trails in the rectified image.
    y_inter_rect : 1d numpy array, float
        Y coordinates of the intersection points of arc lines with
        spectrum trails in the rectified image.
    ttd_order : int or None
        Polynomial order corresponding to the rectification
        transformation. It is None until the coefficients ttd_aij and
        ttd_bij have been computed.
    ttd_aij : numpy array
        Polynomial coefficents corresponding to the direct
        rectification transformation coefficients a_ij.
    ttd_bij : numpy array
        Polynomial coefficents corresponding to the direct
        rectification transformation coefficients b_ij.
    tti_aij : numpy array
        Polynomial coefficents corresponding to the inverse
        rectification transformation coefficients a_ij.
    tti_bij : numpy array
        Polynomial coefficents corresponding to the inverse
        rectification transformation coefficients b_ij.
    ttd_order_longslit_model : int or None
        Polynomial order corresponding to the modeled direct
        rectification transformation. It is None until the coefficients
        ttd_aij_longslit_model and ttd_bij_longslit_model have been
        computed.
    ttd_aij_longslit_model : numpy array
        Polynomial coefficents corresponding to the direct
        rectification transformation coefficients a_ij interpolated
        with a smooth polynomial variation as a function of
        y0_reference_middle.
    ttd_bij_longslit_model : numpy array
        Polynomial coefficents corresponding to the direct
        rectification transformation coefficients b_ij interpolated
        with a smooth polynomial variation as a function of
        y0_reference_middle.
    tti_aij_longslit_model : numpy array
        Polynomial coefficents corresponding to the inverse
        rectification transformation coefficients a_ij interpolated
        with a smooth polynomial variation as a function of
        y0_reference_middle.
    tti_bij_longslit_model : numpy array
        Polynomial coefficents corresponding to the inverse
        rectification transformation coefficients b_ij interpolated
        with a smooth polynomial variation as a function of
        y0_reference_middle.
    wpoly : numpy.polynomial.Polynomial instance
        Wavelength calibration polynomial, providing the
        wavelength as a function of pixel number (running from 1 to
        NAXIS1).
    wpoly_longslit_model : numpy.polynomial.Polynomial instance
        Refined and modeled wavelength calibration polynomial,
        providing the wavelength as a function of pixel number (running
        from 1 to NAXIS1), or None (when the fit cannot be obtained).
    crval1_linear : float
        CRVAL1 corresponding to a linear wavelength scale computed from
        wpoly.
    cdelt1_linear : float
        CDELT1 corresponding to a linear wavelength scale computed from
        wpoly.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    """

    def __init__(self, islitlet,
                 csu_conf, ymargin_bb,
                 params=None, parmodel=None,
                 debugplot=0):

        # slitlet number
        self.islitlet = islitlet

        # debugplot
        self.debugplot = debugplot

        # csu configuration
        self.csu_bar_left = csu_conf.csu_bar_left(islitlet)
        self.csu_bar_right = csu_conf.csu_bar_right(islitlet)
        self.csu_bar_slit_center = csu_conf.csu_bar_slit_center(islitlet)
        self.csu_bar_slit_width = csu_conf.csu_bar_slit_width(islitlet)

        # horizontal bounding box
        self.bb_nc1_orig = 1
        self.bb_nc2_orig = EMIR_NAXIS1

        # reference abscissa
        self.x0_reference = float(EMIR_NAXIS1) / 2.0 + 0.5  # single float

        # define expected frontier ordinates at x0_reference for the rectified
        # image imposing the vertical length of the slitlet to be constant
        # and equal to EMIR_NPIXPERSLIT_RECTIFIED
        self.y0_frontier_lower_expected = expected_y0_lower_frontier(
            islitlet)
        self.y0_frontier_upper_expected = expected_y0_upper_frontier(
            islitlet)

        if params is None:
            return

        # compute spectrum trails and store in which order they are computed
        self.i_lower_spectrail = 0
        self.i_middle_spectrail = 1
        self.i_upper_spectrail = 2
        self.list_spectrails = expected_distorted_boundaries(
                islitlet, self.csu_bar_slit_center,
                [0, 0.5, 1], params, parmodel,
                numpts=101, deg=5, debugplot=0
            )
        # update y_rectified computed at x0_reference
        for spectrail in self.list_spectrails:
            spectrail.y_rectified = spectrail.poly_funct(self.x0_reference)

        # define reference ordinates using lower, middle and upper spectrails
        # evaluated at x0_reference
        self.y0_reference_lower = \
            self.list_spectrails[self.i_lower_spectrail].y_rectified
        self.y0_reference_middle = \
            self.list_spectrails[self.i_middle_spectrail].y_rectified
        self.y0_reference_upper = \
            self.list_spectrails[self.i_upper_spectrail].y_rectified

        # compute frontiers (lower and upper)
        self.list_frontiers = expected_distorted_frontiers(
            islitlet, self.csu_bar_slit_center,
            params, parmodel,
            numpts=101, deg=5, debugplot=0
        )
        # update y_rectified computed at x0_reference
        for spectrail in self.list_frontiers:
            spectrail.y_rectified = spectrail.poly_funct(self.x0_reference)

        # define frontier ordinates at x0_reference
        self.y0_frontier_lower = self.list_frontiers[0].y_rectified
        self.y0_frontier_upper = self.list_frontiers[1].y_rectified

        # determine vertical bounding box
        self.ymargin_bb = ymargin_bb
        xdum = np.linspace(1, EMIR_NAXIS1, num=EMIR_NAXIS1)
        ylower = self.list_frontiers[0].poly_funct(xdum)
        yupper = self.list_frontiers[1].poly_funct(xdum)
        self.bb_ns1_orig = int(ylower.min() + 0.5) - self.ymargin_bb
        if self.bb_ns1_orig < 1:
            self.bb_ns1_orig = 1
        self.bb_ns2_orig = int(yupper.max() + 0.5) + self.ymargin_bb
        if self.bb_ns2_orig > EMIR_NAXIS2:
            self.bb_ns2_orig = EMIR_NAXIS2

        # compute linear transformation to place the rectified slitlet at
        # the center of the current slitlet bounding box
        xdum1 = self.y0_frontier_lower
        ydum1 = self.y0_frontier_lower_expected
        xdum2 = self.y0_frontier_upper
        ydum2 = self.y0_frontier_upper_expected
        self.corr_yrect_b = (ydum2 - ydum1) / (xdum2 - xdum1)
        self.corr_yrect_a = ydum1 - self.corr_yrect_b * xdum1
        # compute expected location of rectified boundaries
        self.y0_reference_lower_expected = \
            self.corr_yrect_a + self.corr_yrect_b * self.y0_reference_lower
        self.y0_reference_middle_expected = \
            self.corr_yrect_a + self.corr_yrect_b * self.y0_reference_middle
        self.y0_reference_upper_expected = \
            self.corr_yrect_a + self.corr_yrect_b * self.y0_reference_upper
        # shift transformation to center the rectified slitlet within the
        # slitlet bounding box
        ydummid = (ydum1 + ydum2) / 2
        ioffset = int(
            ydummid - (self.bb_ns1_orig + self.bb_ns2_orig) / 2.0)
        self.corr_yrect_a -= ioffset

        # minimum and maximum row in the rectified slitlet encompassing
        # EMIR_NPIXPERSLIT_RECTIFIED pixels
        # a) scan number (in pixels, from 1 to NAXIS2)
        xdum1 = self.corr_yrect_a + \
                self.corr_yrect_b * self.y0_frontier_lower
        xdum2 = self.corr_yrect_a + \
                self.corr_yrect_b * self.y0_frontier_upper
        # b) row number (starting from zero)
        self.min_row_rectified = \
            int((round(xdum1 * 10) + 5) / 10) - self.bb_ns1_orig
        self.max_row_rectified = \
            int((round(xdum2 * 10) - 5) / 10) - self.bb_ns1_orig

        # place holder for still undefined class members
        self.list_arc_lines = None
        self.x_inter_orig = None
        self.y_inter_orig = None
        self.x_inter_rect = None
        self.y_inter_rect = None
        self.ttd_order = None
        self.ttd_aij = None
        self.ttd_bij = None
        self.tti_aij = None
        self.tti_bij = None
        self.wpoly = None
        self.crval1_linear = None
        self.cdelt1_linear = None

        # undefined members to be updated with images in longslit mode
        self.ttd_order_longslit_model = None
        self.ttd_aij_longslit_model = None
        self.ttd_bij_longslit_model = None
        self.tti_aij_longslit_model = None
        self.tti_bij_longslit_model = None
        self.wpoly_longslit_model = None

    def __repr__(self):
        """Printable representation of a Slitlet2dArc instance."""

        # string with all the information
        output = "<Slitlet2dArc instance>\n" + \
                 "- islitlet....................: " + \
                 str(self.islitlet) + "\n" + \
                 "- csu_bar_left................: " + \
                 str(self.csu_bar_left) + "\n" + \
                 "- csu_bar_right...............: " + \
                 str(self.csu_bar_right) + "\n" + \
                 "- csu_bar_slit_center.........: " + \
                 str(self.csu_bar_slit_center) + "\n" + \
                 "- csu_bar_slit_width..........: " + \
                 str(self.csu_bar_slit_width) + "\n" + \
                 "- x0_reference................: " + \
                 str(self.x0_reference) + "\n" + \
                 "- y0_reference_lower..........: " + \
                 str(self.y0_reference_lower) + "\n" + \
                 "- y0_reference_middle.........: " + \
                 str(self.y0_reference_middle) + "\n" + \
                 "- y0_reference_upper..........: " + \
                 str(self.y0_reference_upper) + "\n" + \
                 "- y0_frontier_lower..........: " + \
                 str(self.y0_frontier_lower) + "\n" + \
                 "- y0_frontier_upper..........: " + \
                 str(self.y0_frontier_upper) + "\n" + \
                 "- y0_frontier_lower_expected: " + \
                 str(self.y0_frontier_lower_expected) + "\n" + \
                 "- y0_frontier_upper_expected: " + \
                 str(self.y0_frontier_upper_expected) + "\n" + \
                 "- corr_yrect_a................: " + \
                 str(self.corr_yrect_a) + \
                 "- corr_yrect_b................: " + \
                 str(self.corr_yrect_b) + \
                 "- min_row_rectified...........: " + \
                 str(self.min_row_rectified) + \
                 "- max_row_rectified...........: " + \
                 str(self.max_row_rectified) + \
                 "- bb_nc1_orig.................: " + \
                 str(self.bb_nc1_orig) + "\n" + \
                 "- bb_nc2_orig.................: " + \
                 str(self.bb_nc2_orig) + "\n" + \
                 "- bb_ns1_orig.................: " + \
                 str(self.bb_ns1_orig) + "\n" + \
                 "- bb_ns2_orig.................: " + \
                 str(self.bb_ns2_orig) + "\n" + \
                 "- lower spectrail_poly_funct..:\n\t" + \
                 str(self.list_spectrails[self.i_lower_spectrail].poly_funct)\
                 + "\n" + \
                 "- middle spectrail_poly_funct.:\n\t" + \
                 str(self.list_spectrails[self.i_middle_spectrail].poly_funct)\
                 + "\n" + \
                 "- upper spectrail_poly_funct..:\n\t" + \
                 str(self.list_spectrails[self.i_upper_spectrail].poly_funct)\
                 + "\n" + \
                 "- lower frontier_poly_funct...:\n\t" + \
                 str(self.list_frontiers[0].poly_funct) + "\n" + \
                 "- upper frontier_poly_funct...:\n\t" + \
                 str(self.list_frontiers[1].poly_funct) + "\n"

        if self.list_arc_lines is None:
            number_arc_lines = None
        else:
            number_arc_lines = len(self.list_arc_lines)

        output += "- num. of associated arc lines: " + \
                  str(number_arc_lines) + "\n"

        for dumval, dumlab in zip(
                [self.x_inter_orig, self.y_inter_orig,
                 self.x_inter_rect, self.y_inter_rect],
                ["x_inter_orig", "y_inter_orig",
                 "x_inter_rect", "y_inter_rect"]
        ):
            if dumval is None:
                output += "- " + dumlab + "................: None\n"
            else:
                output += "- " + dumlab + "................: " + \
                          str(len(dumval)) + " values defined\n\t[" + \
                          str(dumval[0]) + ", ..., " + \
                          str(dumval[-1]) + "]\n"
        output += \
            "- ttd_order...............: " + str(self.ttd_order) + "\n" + \
            "- ttd_aij.................:\n\t" + str(self.ttd_aij) + "\n" + \
            "- ttd_bij.................:\n\t" + str(self.ttd_bij) + "\n" + \
            "- tti_aij.................:\n\t" + str(self.tti_aij) + "\n" + \
            "- tti_bij.................:\n\t" + str(self.tti_bij) + "\n" + \
            "- ttd_order_longslit_model: " + \
            str(self.ttd_order_longslit_model) + "\n" + \
            "- ttd_aij_longslit_model..:\n\t" + \
            str(self.ttd_aij_longslit_model) + "\n" + \
            "- ttd_bij_longslit_model..:\n\t" + \
            str(self.ttd_bij_longslit_model) + "\n" + \
            "- tti_aij_longslit_model..:\n\t" + \
            str(self.tti_aij_longslit_model) + "\n" + \
            "- tti_bij_longslit_model..:\n\t" + \
            str(self.tti_bij_longslit_model) + "\n" + \
            "- wpoly..................:\n\t" + \
            str(self.wpoly) + "\n" + \
            "- wpoly_longslit_model....:\n\t" + \
            str(self.wpoly_longslit_model) + "\n" + \
            "- crval1_linear...........: " + str(self.crval1_linear) + "\n" + \
            "- cdelt1_linear...........: " + str(self.cdelt1_linear) + "\n" + \
            "- debugplot...............: " + \
            str(self.debugplot)

        return output

    def extract_slitlet2d(self, image_2k2k):
        """Extract slitlet 2d image from image with original EMIR dimensions.

        Parameters
        ----------
        image_2k2k : numpy array
            Original image (dimensions EMIR_NAXIS1 * EMIR_NAXIS2)

        Returns
        -------
        slitlet2d : numpy array
            Image corresponding to the slitlet region defined by its
            bounding box.

        """

        # protections
        naxis2, naxis1 = image_2k2k.shape
        if naxis1 != EMIR_NAXIS1:
            raise ValueError('Unexpected naxis1')
        if naxis2 != EMIR_NAXIS2:
            raise ValueError('Unexpected naxis2')

        # extract slitlet region
        slitlet2d = image_2k2k[(self.bb_ns1_orig - 1):self.bb_ns2_orig,
                               (self.bb_nc1_orig - 1):self.bb_nc2_orig]

        # transform to float
        slitlet2d = slitlet2d.astype(np.float)

        # display slitlet2d with boundaries and middle spectrum trail
        if abs(self.debugplot) in [21, 22]:
            ax = ximshow(slitlet2d, title="Slitlet#" + str(self.islitlet),
                         first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                         show=False)
            xdum = np.linspace(1, EMIR_NAXIS1, num=EMIR_NAXIS1)
            ylower = \
                self.list_spectrails[self.i_lower_spectrail].poly_funct(xdum)
            ax.plot(xdum, ylower, 'b-')
            ymiddle = \
                self.list_spectrails[self.i_middle_spectrail].poly_funct(xdum)
            ax.plot(xdum, ymiddle, 'b--')
            yupper = \
                self.list_spectrails[self.i_upper_spectrail].poly_funct(xdum)
            ax.plot(xdum, yupper, 'b-')
            ylower_frontier = self.list_frontiers[0].poly_funct(xdum)
            ax.plot(xdum, ylower_frontier, 'b:')
            yupper_frontier = self.list_frontiers[1].poly_funct(xdum)
            ax.plot(xdum, yupper_frontier, 'b:')
            pause_debugplot(debugplot=self.debugplot, pltshow=True)

        # return slitlet image
        return slitlet2d

    def locate_unknown_arc_lines(self, slitlet2d,
                                 times_sigma_threshold=15,
                                 minimum_threshold=None,
                                 delta_x_max=30,
                                 delta_y_min=30,
                                 min_dist_from_middle=15):
        """Determine the location of known arc lines in slitlet.

        Parameters
        ----------
        slitlet2d : numpy array
            Image containing the 2d slitlet image.
        times_sigma_threshold : float
            Times (robust) sigma above the median of the image to look
            for arc lines.
        minimum_threshold : float or None
            Minimum threshold to look for arc lines.
        delta_x_max : float
            Maximum size of potential arc line in the X direction.
        delta_y_min : float
            Minimum size of potential arc line in the Y direction.
        min_dist_from_middle : float
            Minimum Y distance from the middle spectrum trail to the
            extreme of the potential arc line. This constraint avoid
            detecting arc line reflections as bone fide arc lines.

        """

        # smooth denoising of slitlet2d
        slitlet2d_rs, coef_rs = rescale_array_to_z1z2(slitlet2d, z1z2=(-1, 1))
        slitlet2d_dn = restoration.denoise_nl_means(slitlet2d_rs,
                                                    patch_size=3,
                                                    patch_distance=2,
                                                    multichannel=False)
        slitlet2d_dn = rescale_array_from_z1z2(slitlet2d_dn, coef_rs)

        # compute basic statistics
        q25, q50, q75 = np.percentile(slitlet2d_dn, q=[25.0, 50.0, 75.0])
        sigmag = 0.7413 * (q75 - q25)  # robust standard deviation
        if abs(self.debugplot) >= 10:
            q16, q84 = np.percentile(slitlet2d_dn, q=[15.87, 84.13])
            print('>>> q16...:', q16)
            print('>>> q25...:', q25)
            print('>>> q50...:', q50)
            print('>>> q75...:', q75)
            print('>>> q84...:', q84)
            print('>>> sigmaG:', sigmag)
        if abs(self.debugplot) in [21, 22]:
            # display initial image with zscale cuts
            title = "Slitlet#" + str(self.islitlet) + \
                    " (locate_unknown_arc_lines, step #1)"
            ximshow(slitlet2d, title=title,
                    first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                    debugplot=self.debugplot)
            # display denoised image with zscale cuts
            title = "Slitlet#" + str(self.islitlet) + \
                    " (locate_unknown_arc_lines, step #2)"
            ximshow(slitlet2d_dn, title=title,
                    first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                    debugplot=self.debugplot)
            # display image with different cuts
            z1z2 = (q50 + times_sigma_threshold * sigmag,
                    q50 + 2 * times_sigma_threshold * sigmag)
            title = "Slitlet#" + str(self.islitlet) + \
                    " (locate_unknown_arc_lines, step #3)"
            ximshow(slitlet2d_dn, title=title, z1z2=z1z2,
                    first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                    debugplot=self.debugplot)

        # determine threshold (using the maximum of q50 + t *sigmag or
        # minimum_threshold)
        threshold = q50 + times_sigma_threshold * sigmag
        if minimum_threshold is not None:
            if minimum_threshold > threshold:
                threshold = minimum_threshold

        # identify objects in slitlet2d above threshold
        labels2d_objects, no_objects = ndimage.label(slitlet2d_dn > threshold)
        if abs(self.debugplot) >= 10:
            print("Number of objects initially found:", no_objects)
        if abs(self.debugplot) in [21, 22]:
            # display all objects identified in the image
            title = "Slitlet#" + str(self.islitlet) + \
                    " (locate_unknown_arc_lines, step #4)"
            z1z2 = (labels2d_objects.min(), labels2d_objects.max())
            ximshow(labels2d_objects, title=title,
                    first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                    cbar_label="Object number",
                    z1z2=z1z2, cmap="nipy_spectral",
                    debugplot=self.debugplot)

        # select arc lines by imposing the criteria based on the
        # dimensions of the detected objects and the intersection with
        # the middle spectrum trail
        slices_possible_arc_lines = ndimage.find_objects(labels2d_objects)
        slices_ok = np.repeat([False], no_objects)  # flag
        for i in range(no_objects):
            if abs(self.debugplot) >= 10:
                print('object', i + 1,
                      '[in np.array coordinates]:',
                      slices_possible_arc_lines[i])
            slice_x = slices_possible_arc_lines[i][1]
            slice_y = slices_possible_arc_lines[i][0]
            # note that the width computation doesn't require to
            # add +1 since slice_x.stop (and slice_y.stop) is
            # already the upper limit +1 (in np.array coordinates)
            delta_x = slice_x.stop - slice_x.start
            delta_y = slice_y.stop - slice_y.start
            # dimensions criterion
            if delta_x <= delta_x_max and delta_y >= delta_y_min:
                # intersection with middle spectrum trail criterion;
                # note that slice_x and slice_y are given in np.array
                # coordinates and are transformed into image coordinates;
                # in addition, -0.5 shift the origin to the lower left
                # corner of the pixel
                xini_slice = slice_x.start + self.bb_nc1_orig - 0.5
                xmiddle_slice = xini_slice + delta_x / 2
                polydum = \
                    self.list_spectrails[self.i_middle_spectrail].poly_funct
                ymiddle_slice = polydum(xmiddle_slice)
                yini_slice = slice_y.start + self.bb_ns1_orig - 0.5
                yend_slice = yini_slice + delta_y
                if yini_slice + min_dist_from_middle <= ymiddle_slice <= \
                        yend_slice - min_dist_from_middle:
                    slices_ok[i] = True

        # generate list with ID of arc lines (note that first object is
        # number 0 and not 1)
        list_slices_ok = []
        for i in range(no_objects):
            if slices_ok[i]:
                list_slices_ok.append(i + 1)
        number_arc_lines = len(list_slices_ok)
        if abs(self.debugplot) >= 10:
            print("\nNumber of arc lines initially identified is:",
                  number_arc_lines)
            if number_arc_lines > 0:
                print("Slice ID of lines passing the selection:\n",
                      list_slices_ok)
        if number_arc_lines == 0:
            return

        # display arc lines
        if abs(self.debugplot) in [21, 22]:
            # display all objects identified in the image
            title = "Slitlet#" + str(self.islitlet) + \
                    " (locate_unknown_arc_lines, step #5)"
            z1z2 = (labels2d_objects.min(),
                    labels2d_objects.max())
            ax = ximshow(labels2d_objects, show=False, title=title,
                         first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                         cbar_label="Object number",
                         z1z2=z1z2, cmap="nipy_spectral",
                         debugplot=self.debugplot)
            # plot rectangle around identified arc lines
            for i in range(no_objects):
                if slices_ok[i]:
                    slice_x = slices_possible_arc_lines[i][1]
                    slice_y = slices_possible_arc_lines[i][0]
                    # note that slice_x and slice_y are given in np.array
                    # coordinates and are transformed into image coordinates;
                    # in addition, -0.5 shift the origin to the lower left
                    # corner of the pixel
                    xini_slice = slice_x.start + self.bb_nc1_orig - 0.5
                    yini_slice = slice_y.start + self.bb_ns1_orig - 0.5
                    # note that the width computation doesn't require to
                    # add +1 since slice_x.stop (and slice_y.stop) is
                    # already the upper limit +1 (in np.array coordinates)
                    xwidth_slice = slice_x.stop - slice_x.start
                    ywidth_slice = slice_y.stop - slice_y.start
                    rect = Rectangle((xini_slice, yini_slice),
                                     xwidth_slice, ywidth_slice,
                                     edgecolor='w', facecolor='none')
                    ax.add_patch(rect)
            # show plot
            pause_debugplot(self.debugplot, pltshow=True)

        # adjust individual arc lines passing the initial selection
        self.list_arc_lines = []  # list of ArcLines
        for k in range(number_arc_lines):  # fit each arc line
            # select points to be fitted for a particular arc line
            xy_tmp = np.where(labels2d_objects == list_slices_ok[k])
            x_tmp = xy_tmp[1] + self.bb_nc1_orig  # use image coordinates
            y_tmp = xy_tmp[0] + self.bb_ns1_orig  # use image coordinates
            w_tmp = slitlet2d_dn[xy_tmp]
            # declare new ArcLine instance
            arc_line = ArcLine()
            # define new ArcLine using a weighted fit
            # (note that it must be X vs Y)
            arc_line.fit(x=x_tmp, y=y_tmp, deg=1, w=w_tmp, y_vs_x=False)
            if len(arc_line.poly_funct.coef) == 2:
                # update list with identified ArcLines
                self.list_arc_lines.append(arc_line)
            else:
                # ignore (sometimes the arc_line.fit returns a constant!)
                pass

        # recompute number_arc_lines just in case in the previous fits
        # some lines have just given a zero degree polynomial
        number_arc_lines = len(self.list_arc_lines)

        # remove arc lines with unexpected slopes
        yfit = np.array([self.list_arc_lines[k].poly_funct.coef[1]
                         for k in range(number_arc_lines)])
        xfit = np.zeros(number_arc_lines)
        # intersection between middle spectrum trail and arc line
        for k in range(number_arc_lines):
            arcline = self.list_arc_lines[k]
            xfit[k], ydum = intersection_spectrail_arcline(
                self.list_spectrails[self.i_middle_spectrail], arcline
            )

        # fit slope versus x-coordinate of the intersection of the arc line
        # with the middle spectrum trail
        if len(yfit) > 5:
            degeff = 5
        else:
            degeff = len(yfit) - 1
        polydum, residum, rejected = polfit_residuals_with_sigma_rejection(
            x=xfit, y=yfit, deg=degeff, times_sigma_reject=4.0,
            xlabel='arc line center (islitlet #' + str(self.islitlet) + ')',
            ylabel='arc line slope', debugplot=0
        )
        # remove rejected arc lines
        if len(rejected) > 0:
            if abs(self.debugplot) >= 10:
                print('Rejecting', sum(rejected),
                      'arc lines with suspicious slopes: Slice ID',
                      [list_slices_ok[k] for k in range(number_arc_lines)
                       if rejected[k]])
            self.list_arc_lines = \
                [self.list_arc_lines[k] for k in range(number_arc_lines)
                 if not rejected[k]]
            # recompute number of arc lines
            number_arc_lines = len(self.list_arc_lines)
            if abs(self.debugplot) >= 10:
                print("\nNumber of arc lines finally identified is:",
                      number_arc_lines)

        if abs(self.debugplot) >= 20:
            # print list of arc lines
            print('\nlist_arc_lines:')
            for k in range(number_arc_lines):
                print(k, '->', self.list_arc_lines[k], '\n')

        # display results
        if abs(self.debugplot) in [21, 22]:
            # generate mask with all the arc-line points passing the selection
            mask_arc_lines = np.zeros_like(slitlet2d_dn)
            for k in list_slices_ok:
                mask_arc_lines[labels2d_objects == k] = 1
            # compute image with only the arc lines passing the selection
            labels2d_arc_lines = labels2d_objects * mask_arc_lines
            # display background image with filtered arc lines
            title = "Slitlet#" + str(self.islitlet) + \
                    " (locate_unknown_arc_lines, step #6)"
            z1z2 = (labels2d_arc_lines.min(),
                    labels2d_arc_lines.max())
            ax = ximshow(labels2d_arc_lines, show=False,
                         first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                         cbar_label="Object number",
                         title=title, z1z2=z1z2, cmap="nipy_spectral",
                         debugplot=self.debugplot)
            # plot weighted fit for each arc line (note that the fit is
            # X vs Y)
            for k in range(number_arc_lines):
                xpol, ypol = self.list_arc_lines[k].linspace_pix()
                ax.plot(xpol, ypol, 'g--')
            # display lower and upper points of each arc line
            x_tmp = [arc_line.xlower_line for arc_line in self.list_arc_lines]
            y_tmp = [arc_line.ylower_line for arc_line in self.list_arc_lines]
            ax.plot(x_tmp, y_tmp, 'w+')
            x_tmp = [arc_line.xupper_line for arc_line in self.list_arc_lines]
            y_tmp = [arc_line.yupper_line for arc_line in self.list_arc_lines]
            ax.plot(x_tmp, y_tmp, 'w+')
            # show plot
            pause_debugplot(self.debugplot, pltshow=True)

    def xy_spectrail_arc_intersections(self, slitlet2d=None):
        """Compute intersection points of spectrum trails with arc lines.

        The member list_arc_lines is updated with new keyword:keyval
        values for each arc line.

        Parameters
        ----------
        slitlet2d : numpy array
            Slitlet image to be displayed with the computed boundaries
            and intersecting points overplotted. This argument is
            optional.

        """

        # protections
        if self.list_arc_lines is None:
            raise ValueError("Arc lines not sought")
        number_spectrum_trails = len(self.list_spectrails)
        if number_spectrum_trails == 0:
            raise ValueError("Number of available spectrum trails is 0")
        number_arc_lines = len(self.list_arc_lines)
        if number_arc_lines == 0:
            raise ValueError("Number of available arc lines is 0")

        # intersection of the arc lines with the spectrum trails
        # (note: the coordinates are computed using pixel values,
        #  ranging from 1 to EMIR_NAXIS1, as given in the original
        #  image reference system ---not in the slitlet image reference
        #  system---)
        self.x_inter_rect = np.array([])  # rectified image coordinates
        self.y_inter_rect = np.array([])  # rectified image coordinates
        for arcline in self.list_arc_lines:
            # middle spectrum trail
            spectrail = self.list_spectrails[self.i_middle_spectrail]
            xroot, yroot = intersection_spectrail_arcline(
                spectrail=spectrail, arcline=arcline
            )
            arcline.x_rectified = xroot
            self.x_inter_rect = np.append(
                self.x_inter_rect, [xroot] * number_spectrum_trails
            )
            for spectrail in self.list_spectrails:
                # compute expected ordinate y_expected in the rectified
                # image
                y_expected = self.corr_yrect_a + self.corr_yrect_b * \
                             spectrail.y_rectified
                self.y_inter_rect = np.append(self.y_inter_rect, y_expected)
        if abs(self.debugplot) >= 10:
            print('>>> y0_frontier_lower_expected........: ',
                  self.y0_frontier_lower_expected)
            print('>>> y0_frontier_upper_expected........: ',
                  self.y0_frontier_upper_expected)
            print('>>> shifted y0_frontier_upper_expected: ',
                  self.corr_yrect_a +
                  self.corr_yrect_b * self.y0_frontier_lower)
            print('>>> shifted y0_frontier_lower_expected: ',
                  self.corr_yrect_a +
                  self.corr_yrect_b * self.y0_frontier_upper)
        #
        self.x_inter_orig = np.array([])  # original image coordinates
        self.y_inter_orig = np.array([])  # original image coordinates
        for arcline in self.list_arc_lines:
            for spectrail in self.list_spectrails:
                xroot, yroot = intersection_spectrail_arcline(
                    spectrail=spectrail, arcline=arcline
                )
                self.x_inter_orig = np.append(self.x_inter_orig, xroot)
                self.y_inter_orig = np.append(self.y_inter_orig, yroot)

        # display intersection points
        if abs(self.debugplot % 10) != 0 and slitlet2d is not None:
            # display image with zscale cuts
            title = "Slitlet#" + str(self.islitlet) + \
                    " (xy_spectrail_arc_intersections)"
            ax = ximshow(slitlet2d, title=title,
                         first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                         show=False)
            # spectrum trails
            for spectrail in self.list_spectrails:
                xdum, ydum = spectrail.linspace_pix(start=self.bb_nc1_orig,
                                                    stop=self.bb_nc2_orig)
                ax.plot(xdum, ydum, 'g')
            # arc lines
            for arcline in self.list_arc_lines:
                xdum, ydum = arcline.linspace_pix(start=self.bb_ns1_orig,
                                                  stop=self.bb_ns2_orig)
                ax.plot(xdum, ydum, 'g')
            # intersection points
            ax.plot(self.x_inter_orig, self.y_inter_orig, 'co')
            ax.plot(self.x_inter_rect, self.y_inter_rect, 'bo')
            # show plot
            pause_debugplot(self.debugplot, pltshow=True)

    def estimate_tt_to_rectify(self, order, slitlet2d=None):
        """Estimate the polynomial transformation to rectify the image.

        Parameters
        ----------
        order : int
            Order of the polynomial transformation.
        slitlet2d : numpy array
            Slitlet image to be displayed with the computed boundaries
            and intersecting points overplotted. This argument is
            optional.

        """

        # protections
        if self.x_inter_orig is None \
                or self.y_inter_orig is None \
                or self.x_inter_rect is None \
                or self.y_inter_rect is None:
            raise ValueError('Intersection points not computed')

        npoints = len(self.x_inter_orig)
        if len(self.y_inter_orig) != npoints \
                or len(self.x_inter_rect) != npoints \
                or len(self.y_inter_rect) != npoints:
            raise ValueError('Unexpected different number of points')

        # IMPORTANT: correct coordinates from origin in order to manipulate
        # coordinates corresponding to image indices
        x_inter_orig_shifted = self.x_inter_orig - self.bb_nc1_orig
        y_inter_orig_shifted = self.y_inter_orig - self.bb_ns1_orig
        x_inter_rect_shifted = self.x_inter_rect - self.bb_nc1_orig
        y_inter_rect_shifted = self.y_inter_rect - self.bb_ns1_orig

        # compute 2D transformation
        self.ttd_order = order
        self.ttd_aij, self.ttd_bij = compute_distortion(
            x_inter_orig_shifted, y_inter_orig_shifted,
            x_inter_rect_shifted, y_inter_rect_shifted,
            order,
            self.debugplot
        )
        self.tti_aij, self.tti_bij = compute_distortion(
            x_inter_rect_shifted, y_inter_rect_shifted,
            x_inter_orig_shifted, y_inter_orig_shifted,
            order,
            self.debugplot
        )

        # display slitlet with intersection points and grid indicating
        # the fitted transformation
        if abs(self.debugplot % 10) != 0 and slitlet2d is not None:
            # display image with zscale cuts
            title = "Slitlet#" + str(self.islitlet) + \
                    " (estimate_tt_to_rectify)"
            ax = ximshow(slitlet2d, title=title,
                         first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                         show=False)
            # intersection points
            ax.plot(self.x_inter_orig, self.y_inter_orig, 'co')
            ax.plot(self.x_inter_rect, self.y_inter_rect, 'bo')
            # grid with fitted transformation: spectrum trails
            xx = np.arange(0, self.bb_nc2_orig - self.bb_nc1_orig + 1,
                           dtype=np.float)
            for spectrail in self.list_spectrails:
                yy0 = self.corr_yrect_a + \
                      self.corr_yrect_b * spectrail.y_rectified
                yy = np.tile([yy0 - self.bb_ns1_orig], xx.size)
                ax.plot(xx + self.bb_nc1_orig, yy + self.bb_ns1_orig, "b")
                xxx, yyy = fmap(self.ttd_order, self.ttd_aij, self.ttd_bij,
                                xx, yy)
                ax.plot(xxx + self.bb_nc1_orig, yyy + self.bb_ns1_orig, "g")
            # grid with fitted transformation: arc lines
            ylower_line = \
                self.list_spectrails[self.i_lower_spectrail].y_rectified
            ylower_line = self.corr_yrect_a + self.corr_yrect_b * ylower_line
            yupper_line = \
                self.list_spectrails[self.i_upper_spectrail].y_rectified
            yupper_line = self.corr_yrect_a + self.corr_yrect_b * yupper_line
            n_points = int(yupper_line - ylower_line + 0.5) + 1
            yy = np.linspace(ylower_line - self.bb_ns1_orig,
                             yupper_line - self.bb_ns1_orig,
                             num=n_points,
                             dtype=np.float)
            for arc_line in self.list_arc_lines:
                xline = arc_line.x_rectified - self.bb_nc1_orig
                xx = np.array([xline] * n_points)
                ax.plot(xx + self.bb_nc1_orig,yy + self.bb_ns1_orig, "b" )
                xxx, yyy = fmap(self.ttd_order, self.ttd_aij, self.ttd_bij,
                                xx, yy)
                ax.plot(xxx + self.bb_nc1_orig, yyy + self.bb_ns1_orig, "c")
            # show plot
            pause_debugplot(self.debugplot, pltshow=True)

    def rectify(self, slitlet2d, resampling, transformation, inverse=False):
        """Rectify slitlet using computed transformation.

        Parameters
        ----------
        slitlet2d : numpy array
            Image containing the 2d slitlet image.
        resampling : int
            1: nearest neighbour, 2: flux preserving interpolation.
        transformation : int
            1: initial, 2: modeled
        inverse : bool
            If true, the inverse rectification transformation is
            employed.

        Returns
        -------
        slitlet2d_rect : numpy array
            Rectified slitlet image.

        """

        if resampling not in [1, 2]:
            raise ValueError("Unexpected resampling value=" + str(resampling))

        if transformation not in [1, 2]:
            raise ValueError("Unexpected transformation value=" +
                             str(transformation))

        # verify image dimension
        naxis2, naxis1 = slitlet2d.shape
        if naxis1 != self.bb_nc2_orig - self.bb_nc1_orig + 1:
            raise ValueError("Unexpected slitlet2d_rect naxis1")
        if naxis2 != self.bb_ns2_orig - self.bb_ns1_orig + 1:
            raise ValueError("Unexpected slitlet2d_rect naxis2")

        # transformation to be employed (and direction)
        if transformation == 1:
            if inverse:
                aij = self.tti_aij
                bij = self.tti_bij
            else:
                aij = self.ttd_aij
                bij = self.ttd_bij
        else:
            if inverse:
                aij = self.tti_aij_longslit_model
                bij = self.tti_bij_longslit_model
            else:
                aij = self.ttd_aij_longslit_model
                bij = self.ttd_bij_longslit_model

        # rectify image
        slitlet2d_rect = rectify2d(
            image2d=slitlet2d,
            aij=aij,
            bij=bij,
            resampling=resampling
        )

        if abs(self.debugplot % 10) != 0:
            title = "Slitlet#" + str(self.islitlet) + " (rectify)"
            ax = ximshow(slitlet2d_rect, title=title,
                         first_pixel=(self.bb_nc1_orig, self.bb_ns1_orig),
                         show=False)
            if self.list_arc_lines is not None:
                # intersection points
                ax.plot(self.x_inter_rect, self.y_inter_rect, 'bo')
            # grid with fitted transformation: spectrum trails
            xx = np.arange(0, self.bb_nc2_orig - self.bb_nc1_orig + 1,
                           dtype=np.float)
            for spectrail in self.list_spectrails:
                yy0 = self.corr_yrect_a + \
                      self.corr_yrect_b * spectrail.y_rectified
                yy = np.tile([yy0 - self.bb_ns1_orig], xx.size)
                ax.plot(xx + self.bb_nc1_orig, yy + self.bb_ns1_orig, "b")
            for spectrail in self.list_frontiers:
                yy0 = self.corr_yrect_a + \
                      self.corr_yrect_b * spectrail.y_rectified
                yy = np.tile([yy0 - self.bb_ns1_orig], xx.size)
                ax.plot(xx + self.bb_nc1_orig, yy + self.bb_ns1_orig, "b:")
            # grid with fitted transformation: arc lines
            ylower_line = \
                self.list_spectrails[self.i_lower_spectrail].y_rectified
            ylower_line = self.corr_yrect_a + self.corr_yrect_b * ylower_line
            yupper_line = \
                self.list_spectrails[self.i_upper_spectrail].y_rectified
            yupper_line = self.corr_yrect_a + self.corr_yrect_b * yupper_line
            n_points = int(yupper_line - ylower_line + 0.5) + 1
            yy = np.linspace(ylower_line - self.bb_ns1_orig,
                             yupper_line - self.bb_ns1_orig,
                             num=n_points,
                             dtype=np.float)
            if self.list_arc_lines is not None:
                for arc_line in self.list_arc_lines:
                    xline = arc_line.x_rectified - self.bb_nc1_orig
                    xx = np.array([xline] * n_points)
                    ax.plot(xx + self.bb_nc1_orig, yy + self.bb_ns1_orig, "b")
            # show plot
            pause_debugplot(self.debugplot, pltshow=True)

        return slitlet2d_rect

    def median_spectrum_from_rectified_image(self, slitlet2d_rect,
                                             sigma_gaussian_filtering=0,
                                             nwinwidth_initial=5,
                                             nwinwidth_refined=7,
                                             times_sigma_threshold=5,
                                             minimum_threshold=None,
                                             npix_avoid_border=0,
                                             nbrightlines=None):
        """Median spectrum and line peaks from rectified image.

        In order to avoid the line ghosts, the line peaks are identified
        independently in the upper and lower halves of the rectified
        image. The final peaks correspond to lines that appear in both
        spectra.

        Parameters
        ----------
        slitlet2d_rect : numpy array
            Rectified slitlet image.
        sigma_gaussian_filtering : float
            Sigma of the gaussian filter to be applied to the spectrum
            in order to avoid problems with saturated lines. This
            filtering is skipped when this parameter is <= 0.
        nwinwidth_initial : int
            Width of the window where each peak must be found using
            the initial method (approximate)
        nwinwidth_refined : int
            Width of the window where each peak location will be
            refined.
        times_sigma_threshold : float
            Times (robust) sigma above the median of the image to set
            the minimum threshold when searching for line peaks.
        minimum_threshold : float or None
            Minimum value of the threshold.
        npix_avoid_border : int
            Number of pixels at the borders of the spectrum where peaks
            are not considered. If zero, the actual number will be
            given by nwinwidth_initial.
        nbrightlines : list
            List with maximum number of brightest lines to be employed
            in the wavelength calibration. The length of the list
            indicates the number of equal-size subintervals in the
            wavelength calibration direction to be considered.
            If this value is [0] or None, all the detected lines will be
            employed.

        Returns
        -------
        sp0 : numpy array
            Median spectrum.
        fxpeaks : numpy array
            Refined location of arc lines (in array index scale).

        """

        # protections
        naxis2, naxis1 = slitlet2d_rect.shape
        if naxis1 != self.bb_nc2_orig - self.bb_nc1_orig + 1:
            raise ValueError("Unexpected slitlet2d_rect naxis1")
        if naxis2 != self.bb_ns2_orig - self.bb_ns1_orig + 1:
            raise ValueError("Unexpected slitlet2d_rect naxis2")

        # lower, middle and upper spectrum trails
        ylower_line = \
            self.list_spectrails[self.i_lower_spectrail].y_rectified
        ymiddle_line = \
            self.list_spectrails[self.i_middle_spectrail].y_rectified
        yupper_line = \
            self.list_spectrails[self.i_upper_spectrail].y_rectified

        ilower = int(ylower_line + 0.5) - self.bb_ns1_orig
        imiddle = int(ymiddle_line + 0.5) - self.bb_ns1_orig
        iupper = int(yupper_line + 0.5) - self.bb_ns1_orig

        # median spectra using different image regions
        sp0_ini = np.median(slitlet2d_rect[ilower:(iupper + 1), :], axis=0)
        sp1_ini = np.median(slitlet2d_rect[ilower:(imiddle + 1), :], axis=0)
        sp2_ini = np.median(slitlet2d_rect[imiddle:(iupper + 1), :], axis=0)

        # gaussian filtering when requested (to avoid line saturation)
        if sigma_gaussian_filtering > 0:
            sp0 = ndimage.filters.gaussian_filter(
                sp0_ini,
                sigma=sigma_gaussian_filtering
            )
            sp1 = ndimage.filters.gaussian_filter(
                sp1_ini,
                sigma=sigma_gaussian_filtering
            )
            sp2 = ndimage.filters.gaussian_filter(
                sp2_ini,
                sigma=sigma_gaussian_filtering
            )
        else:
            sp0 = np.copy(sp0_ini)
            sp1 = np.copy(sp1_ini)
            sp2 = np.copy(sp2_ini)

        # compute threshold
        q25, q50, q75 = np.percentile(sp0, q=[25.0, 50.0, 75.0])
        sigma_g = 0.7413 * (q75 - q25)  # robust standard deviation
        threshold = q50 + times_sigma_threshold * sigma_g
        if abs(self.debugplot) >= 10:
            print("median...........:", q50)
            print("robuts std.......:", sigma_g)
            print("threshold........:", threshold)
        if minimum_threshold is not None:
            if minimum_threshold > threshold:
                threshold = minimum_threshold
        if abs(self.debugplot) >= 10:
            print("minimum threshold:", minimum_threshold)
            print("final threshold..:", threshold)

        # initial location of the peaks (integer values)
        ixpeaks0 = find_peaks_spectrum(sp0, nwinwidth=nwinwidth_initial,
                                       threshold=threshold,
                                       debugplot=self.debugplot)

        # peaks in the lower and upper regions
        ixpeaks1 = find_peaks_spectrum(sp1, nwinwidth=nwinwidth_initial,
                                       threshold=threshold,
                                       debugplot=self.debugplot)
        ixpeaks2 = find_peaks_spectrum(sp2, nwinwidth=nwinwidth_initial,
                                       threshold=threshold,
                                       debugplot=self.debugplot)

        # the peaks are valid if the are also found in the lower and
        # upper regions (with a tolerance of +1 or -1 pixel)
        ixpeaks = []
        for ixpeak in ixpeaks0:
            l1 = ixpeak in np.concatenate((ixpeaks1, ixpeaks1+1, ixpeaks1-1))
            l2 = ixpeak in np.concatenate((ixpeaks2, ixpeaks2+1, ixpeaks2-1))
            if l1 and l2:
                ixpeaks.append(ixpeak)
        ixpeaks = np.array(ixpeaks)
        if abs(self.debugplot) >= 10:
            print("Merged initial list of peaks:\n", ixpeaks)

        # remove peaks too close to any of the borders of the spectrum
        if npix_avoid_border > 0:
            lok_ini = ixpeaks >= npix_avoid_border
            lok_end = ixpeaks <= len(sp0) - 1 - npix_avoid_border
            ixpeaks = ixpeaks[lok_ini * lok_end]

        # select a maximum number of brightest lines in each region
        if nbrightlines is None:
            pass
        elif len(nbrightlines) == 1 and nbrightlines[0] == 0:
            pass
        else:
            if abs(self.debugplot) >= 10:
                print('nbrightlines =', nbrightlines)
                print('ixpeaks in whole spectrum:\n', ixpeaks)
            region_size = (naxis1-1)/len(nbrightlines)
            ixpeaks_filtered = np.array([], dtype=int)
            for iregion, nlines_in_region in enumerate(nbrightlines):
                if nlines_in_region > 0:
                    imin = int(iregion * region_size)
                    imax = int((iregion + 1) * region_size)
                    if iregion > 0:
                        imin += 1
                    ixpeaks_region = \
                        ixpeaks[np.logical_and(ixpeaks >= imin,
                                               ixpeaks <= imax)]
                    if len(ixpeaks_region) > 0:
                        peak_fluxes = sp0[ixpeaks_region]
                        spos = peak_fluxes.argsort()
                        ixpeaks_tmp = ixpeaks_region[spos[-nlines_in_region:]]
                        ixpeaks_tmp.sort()  # in-place sort
                        if abs(self.debugplot) >= 10:
                            print('ixpeaks in region........:\n', ixpeaks_tmp)
                        ixpeaks_filtered = np.concatenate((ixpeaks_filtered,
                                                           ixpeaks_tmp))
            ixpeaks = ixpeaks_filtered
            if abs(self.debugplot) >= 10:
                print('ixpeaks filtered.........:\n', ixpeaks)

        # refined location of the peaks (float values)
        fxpeaks, sxpeaks = refine_peaks_spectrum(sp0, ixpeaks,
                                                 nwinwidth=nwinwidth_refined,
                                                 method="gaussian")

        if abs(self.debugplot) % 10 != 0:
            x = np.arange(self.bb_nc1_orig, self.bb_nc2_orig + 1)
            title = "Slitlet#" + str(self.islitlet) + " (median spectrum)"
            ax = ximplotxy(x, sp1, show=False, title=title,
                           xlabel='pixel coordinate (from 1 to NAXIS1)',
                           ylabel='number of counts',
                           **{'marker': ' ', 'label': 'lower region'})
            ax.plot(x, sp2, label='upper region')
            ax.plot(x, sp0, label='whole region')
            # mark peak location
            ax.plot(ixpeaks + self.bb_nc1_orig,
                    sp0[ixpeaks], 'o', label="initial location")
            ax.plot(fxpeaks + self.bb_nc1_orig,
                    sp0[ixpeaks], 'o', label="refined location")
            ax.legend()
            pause_debugplot(self.debugplot, pltshow=True, tight_layout=False)

        # return median spectrum and refined peak location
        return sp0, fxpeaks



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
from copy import deepcopy
from datetime import datetime
import json
from lmfit import Minimizer, Parameters, report_fit
import matplotlib.pyplot as plt
import numpy as np
import pickle
import sys
from uuid import uuid4

from numina.array.ccd_line import SpectrumTrail
from numina.array.display.matplotlib_qt import set_window_geometry
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximshow import ximshow
from emirdrp.core import EMIR_NBARS
from emirdrp.instrument.dtu_configuration import DtuConfiguration
from emirdrp.instrument.dtu_configuration import average_dtu_configurations
from emirdrp.instrument.dtu_configuration import maxdiff_dtu_configurations


from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from numina.array.display.pause_debugplot import DEBUGPLOT_CODES
from numina.tools.arg_file_is_new import arg_file_is_new

EXPECTED_PARAMETER_LIST = (
    'c2', 'c4', 'ff', 'slit_gap', 'slit_height',
    'theta0_origin', 'theta0_slope', 'x0', 'y0', 'y_baseline'
)

EXPECTED_PARAMETER_LIST_EXTENDED = tuple(
    mainpar + '_' + subpar for mainpar in EXPECTED_PARAMETER_LIST
    for subpar in ['a0s', 'a1s', 'a2s']
)

FUNCTION_EVALUATIONS = 0


def integrity_check(bounddict, max_dtu_offset):
    """Integrity check of 'bounddict' content.

    Parameters
    ----------
    bounddict : JSON structure
        Structure employed to store bounddict information.
    max_dtu_offset : float
        Maximum allowed difference in DTU location (mm) for each
        parameter

    """

    if 'meta_info' not in bounddict.keys():
        raise ValueError('"meta_info" not found in JSON file')
    if 'description' not in bounddict['meta_info'].keys():
        raise ValueError('"description" not found in JSON file')
    if bounddict['meta_info']['description'] != \
            'slitlet boundaries from fits to continuum-lamp exposures':
        raise ValueError('Unexpected "description" in JSON file')

    grism = bounddict['tags']['grism']
    print('>>> grism...:', grism)
    spfilter = bounddict['tags']['filter']
    print('>>> filter..:', spfilter)

    valid_slitlets = ["slitlet" + str(i).zfill(2) for i in
                      range(1, EMIR_NBARS + 1)]
    read_slitlets = list(bounddict['contents'].keys())
    read_slitlets.sort()

    first_dtu = True
    first_dtu_configuration = None  # avoid PyCharm warning
    list_dtu_configurations = []

    for tmp_slitlet in read_slitlets:
        if tmp_slitlet not in valid_slitlets:
            raise ValueError("Unexpected slitlet key: " + tmp_slitlet)
        # for each slitlet, check valid DATE-OBS (ISO 8601)
        read_dateobs = list(bounddict['contents'][tmp_slitlet].keys())
        read_dateobs.sort()
        for tmp_dateobs in read_dateobs:
            try:
                datetime.strptime(tmp_dateobs, "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError:
                print("Unexpected date_obs key: " + tmp_dateobs)
                raise
            # for each DATE-OBS, check expected fields
            tmp_dict = bounddict['contents'][tmp_slitlet][tmp_dateobs]
            valid_keys = ["boundary_coef_lower",
                          "boundary_coef_upper",
                          "boundary_xmax_lower",
                          "boundary_xmax_upper",
                          "boundary_xmin_lower",
                          "boundary_xmin_upper",
                          "csu_bar_left",
                          "csu_bar_right",
                          "csu_bar_slit_center",
                          "csu_bar_slit_width",
                          "rotang",
                          "xdtu",
                          "xdtu_0",
                          "ydtu",
                          "ydtu_0",
                          "zdtu",
                          "zdtu_0",
                          "zzz_info1",
                          "zzz_info2"]
            read_keys = tmp_dict.keys()
            for tmp_key in read_keys:
                if tmp_key not in valid_keys:
                    print("ERROR:")
                    print("grism...:", grism)
                    print("slitlet.:", tmp_slitlet)
                    print("date_obs:", tmp_dateobs)
                    raise ValueError("Unexpected key " + tmp_key)
            for tmp_key in valid_keys:
                if tmp_key not in read_keys:
                    print("ERROR:")
                    print("grism...:", grism)
                    print("slitlet.:", tmp_slitlet)
                    print("date_obs:", tmp_dateobs)
                    raise ValueError("Expected key " + tmp_key + " not found")
            if tmp_dict['boundary_xmax_lower'] <= \
                    tmp_dict['boundary_xmin_lower']:
                print("ERROR:")
                print("grism...:", grism)
                print("slitlet.:", tmp_slitlet)
                print("date_obs:", tmp_dateobs)
                print("boundary_xmin_lower", tmp_dict['boundary_xmin_lower'])
                print("boundary_xmax_lower", tmp_dict['boundary_xmax_lower'])
                raise ValueError("Unexpected boundary_xmax_lower <= "
                                 "boundary_xmin_lower")
            if tmp_dict['boundary_xmax_upper'] <= \
                    tmp_dict['boundary_xmin_upper']:
                print("ERROR:")
                print("grism...:", grism)
                print("slitlet.:", tmp_slitlet)
                print("date_obs:", tmp_dateobs)
                print("boundary_xmin_upper", tmp_dict['boundary_xmin_upper'])
                print("boundary_xmax_upper", tmp_dict['boundary_xmax_upper'])
                raise ValueError("Unexpected boundary_xmax_upper <= "
                                 "boundary_xmin_upper")
            if first_dtu:
                first_dtu_configuration = \
                    DtuConfiguration.define_from_dictionary(tmp_dict)
                first_dtu = False
                list_dtu_configurations.append(first_dtu_configuration)
            else:
                last_dtu_configuration = \
                    DtuConfiguration.define_from_dictionary(tmp_dict)
                if not first_dtu_configuration.closeto(
                        last_dtu_configuration,
                        abserror=max_dtu_offset
                ):
                    print("ERROR:")
                    print("grism...:", grism)
                    print("slitlet.:", tmp_slitlet)
                    print("date_obs:", tmp_dateobs)
                    print("First DTU configuration..:\n\t",
                          first_dtu_configuration)
                    print("Last DTU configuration...:\n\t",
                          last_dtu_configuration)
                    raise ValueError("Unexpected DTU configuration")
                list_dtu_configurations.append(last_dtu_configuration)

    print("* Integrity check OK!")

    averaged_dtu_configuration = average_dtu_configurations(
        list_dtu_configurations)
    maxdiff_dtu_configuration = maxdiff_dtu_configurations(
        list_dtu_configurations
    )

    return averaged_dtu_configuration, maxdiff_dtu_configuration


def exvp_scalar(x, y, x0, y0, c2, c4, theta0, ff):
    """Convert virtual pixel to real pixel.

    Parameters
    ----------
    x : array-like of floats
        X coordinate (pixel).
    y : array-like of floats
        Y coordinate (pixel).
    x0 : float
        X coordinate of reference pixel, in units of 1E3.
    y0 : float
        Y coordinate of reference pixel, in units of 1E3.
    c2 : float
        Coefficient corresponding to the term r**2 in distortion
        equation, in units of 1E4.
    c4 : float
        Coefficient corresponding to the term r**4 in distortion
        equation, in units of 1E9
    theta0 : float
        Additional rotation angle (radians).
    ff : float
        Scaling factor to be applied to the Y axis.

    Returns
    -------
    xdist, ydist : tuple of floats
        Distorted coordinates.

    """

    # plate scale: 0.1944 arcsec/pixel
    # conversion factor (in radian/pixel)
    factor = 0.1944 * np.pi/(180.0*3600)
    # distance from image center (pixels)
    r_pix = np.sqrt((x - x0*1000)**2 + (y - y0*1000)**2)
    # distance from imagen center (radians)
    r_rad = factor * r_pix
    # radial distortion: this number is 1.0 for r=0 and increases
    # slightly (reaching values around 1.033) for r~sqrt(2)*1024
    # (the distance to the corner of the detector measured from the
    # center)
    rdist = (1 +
             c2 * 1.0E4 * r_rad**2 +
             c4 * 1.0E9 * r_rad**4)
    # angle measured from the Y axis towards the X axis
    theta = np.arctan((x - x0*1000)/(y - y0*1000))
    if y < y0*1000:
        theta = theta - np.pi
    # distorted coordinates
    xdist = (rdist * r_pix * np.sin(theta+theta0)) + x0*1000
    ydist = (ff * rdist * r_pix * np.cos(theta+theta0)) + y0*1000

    return xdist, ydist


def exvp(x, y, x0, y0, c2, c4, theta0, ff):
    """Convert virtual pixel(s) to real pixel(s).

    This function makes use of exvp_scalar(), which performs the
    conversion for a single point (x, y), over an array of X and Y
    values.

    Parameters
    ----------
    x : array-like
        X coordinate (pixel).
    y : array-like
        Y coordinate (pixel).
    x0 : float
        X coordinate of reference pixel.
    y0 : float
        Y coordinate of reference pixel.
    c2 : float
        Coefficient corresponding to the term r**2 in distortion
        equation.
    c4 : float
        Coefficient corresponding to the term r**4 in distortion
        equation.
    theta0 : float
        Additional rotation angle (radians).
    ff : float
        Scaling factor to be applied to the Y axis.

    Returns
    -------
    xdist, ydist : tuple of floats (or two arrays of floats)
        Distorted coordinates.

    """

    if all([np.isscalar(x), np.isscalar(y)]):
        xdist, ydist = exvp_scalar(x, y, x0=x0, y0=y0,
                                   c2=c2, c4=c4, theta0=theta0, ff=ff)
        return xdist, ydist
    elif any([np.isscalar(x), np.isscalar(y)]):
        raise ValueError("invalid mixture of scalars and arrays")
    else:
        xdist = []
        ydist = []
        for x_, y_ in zip(x, y):
            xdist_, ydist_ = exvp_scalar(x_, y_, x0=x0, y0=y0,
                                         c2=c2, c4=c4, theta0=theta0, ff=ff)
            xdist.append(xdist_)
            ydist.append(ydist_)
        return np.array(xdist), np.array(ydist)


def return_params(islitlet, csu_bar_slit_center, params, parmodel):
    """Return individual model parameters from object of type Parameters.

    Parameters
    ----------
    islitlet : int
        Number of slitlet.
    csu_bar_slit_center : float
        CSU bar slit center, in mm.
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.


    Returns
    -------
    c2 : float
        Coefficient corresponding to the term r**2 in distortion
        equation.
    c4 : float
        Coefficient corresponding to the term r**4 in distortion
        equation.
    ff : float
        Scaling factor to be applied to the Y axis.
    slit_gap : float
        Slit gap.
    slit_height : float
        Slit height.
    theta0 : float
        Additional rotation angle (radians).
    x0 : float
        X coordinate of reference pixel.
    y0 : float
        Y coordinate of reference pixel.
    y_baseline : float
        Y coordinate employed as baseline.

    """

    if parmodel == "longslit":
        # set each variable in EXPECTED_PARAMETER_LIST to the value
        # transferred through 'params'
        c2 = params['c2'].value
        c4 = params['c4'].value
        ff = params['ff'].value
        slit_gap = params['slit_gap'].value
        slit_height = params['slit_height'].value
        theta0_origin = params['theta0_origin'].value
        theta0_slope = params['theta0_slope'].value
        x0 = params['x0'].value
        y0 = params['y0'].value
        y_baseline = params['y_baseline'].value
    else:
        # set each variable in EXPECTED_PARAMETER_LIST_EXTENDED to the value
        # transferred through 'params'
        c2_a0s = params['c2_a0s'].value
        c2_a1s = params['c2_a1s'].value / 1E3
        c2_a2s = params['c2_a2s'].value / 1E6
        c2 = c2_a0s + \
             c2_a1s * csu_bar_slit_center + \
             c2_a2s * csu_bar_slit_center ** 2
        # ---
        c4_a0s = params['c4_a0s'].value
        c4_a1s = params['c4_a1s'].value / 1E3
        c4_a2s = params['c4_a2s'].value / 1E6
        c4 = c4_a0s + \
             c4_a1s * csu_bar_slit_center + \
             c4_a2s * csu_bar_slit_center ** 2
        # ---
        ff_a0s = params['ff_a0s'].value
        ff_a1s = params['ff_a1s'].value / 1E3
        ff_a2s = params['ff_a2s'].value / 1E6
        ff = ff_a0s + \
             ff_a1s * csu_bar_slit_center + \
             ff_a2s * csu_bar_slit_center ** 2
        # ---
        slit_gap_a0s = params['slit_gap_a0s'].value
        slit_gap_a1s = params['slit_gap_a1s'].value / 1E3
        slit_gap_a2s = params['slit_gap_a2s'].value / 1E6
        slit_gap = slit_gap_a0s + \
                   slit_gap_a1s * csu_bar_slit_center + \
                   slit_gap_a2s * csu_bar_slit_center ** 2
        # ---
        slit_height_a0s = params['slit_height_a0s'].value
        slit_height_a1s = params['slit_height_a1s'].value / 1E3
        slit_height_a2s = params['slit_height_a2s'].value / 1E6
        slit_height = slit_height_a0s + \
                      slit_height_a1s * csu_bar_slit_center + \
                      slit_height_a2s * csu_bar_slit_center ** 2
        # ---
        theta0_origin_a0s = params['theta0_origin_a0s'].value
        theta0_origin_a1s = params['theta0_origin_a1s'].value / 1E3
        theta0_origin_a2s = params['theta0_origin_a2s'].value / 1E6
        theta0_origin = theta0_origin_a0s + \
                        theta0_origin_a1s * csu_bar_slit_center + \
                        theta0_origin_a2s * csu_bar_slit_center ** 2
        # ---
        theta0_slope_a0s = params['theta0_slope_a0s'].value
        theta0_slope_a1s = params['theta0_slope_a1s'].value / 1E3
        theta0_slope_a2s = params['theta0_slope_a2s'].value / 1E6
        theta0_slope = theta0_slope_a0s + \
                       theta0_slope_a1s * csu_bar_slit_center + \
                       theta0_slope_a2s * csu_bar_slit_center ** 2
        # ---
        x0_a0s = params['x0_a0s'].value
        x0_a1s = params['x0_a1s'].value / 1E3
        x0_a2s = params['x0_a2s'].value / 1E6
        x0 = x0_a0s + \
             x0_a1s * csu_bar_slit_center + \
             x0_a2s * csu_bar_slit_center ** 2
        # ---
        y0_a0s = params['y0_a0s'].value
        y0_a1s = params['y0_a1s'].value / 1E3
        y0_a2s = params['y0_a2s'].value / 1E6
        y0 = y0_a0s + \
             y0_a1s * csu_bar_slit_center + \
             y0_a2s * csu_bar_slit_center ** 2
        # ---
        y_baseline_a0s = params['y_baseline_a0s'].value
        y_baseline_a1s = params['y_baseline_a1s'].value / 1E3
        y_baseline_a2s = params['y_baseline_a2s'].value / 1E6
        y_baseline = y_baseline_a0s + \
                     y_baseline_a1s * csu_bar_slit_center + \
                     y_baseline_a2s * csu_bar_slit_center ** 2

    theta0 = theta0_origin / 1E3 + theta0_slope / 1E4 * islitlet

    return c2, c4, ff, slit_gap, slit_height, theta0, x0, y0, y_baseline


def expected_distorted_boundaries(islitlet, csu_bar_slit_center,
                                  borderlist, params, parmodel,
                                  numpts, deg, debugplot=0):
    """Return expected SpectrumTrail instances associated to a given slitlet.

    Several SpectrumTrail objects can be computed for the considered
    slitlet. The parameter borderlist is a list of floats, ranging
    from 0 to 1, indicating the spatial location of the spectrum trail
    within the slitlet: 0 means the lower boundary and 1 corresponds
    to the upper bounday. Any value in (0,1) will provide the
    spectrum trail located in between accordingly.

    Parameters
    ----------
    islitlet : int
        Number of slitlet.
    csu_bar_slit_center : float
        CSU bar slit center, in mm.
    borderlist : list of floats
        Each float provides the fractional vertical location of the
        spectrum trail relative to the lower boundary. In other words,
        0.0 corresponds to the lower boundary, 1.0 to the upper
        boundary, and any number in the interval (0,1) will be a
        spectral trail in between.
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    numpts : int
        Number of points in which the X-range interval is subdivided
        before fitting the returned polynomial(s).
    deg : int
        Degree of the fitted polynomial.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    list_spectrails : list of SpectrumTrail objects
        List containing the fitted spectrum trails.

    """

    c2, c4, ff, slit_gap, slit_height, theta0, x0, y0, y_baseline = \
        return_params(islitlet, csu_bar_slit_center, params, parmodel)

    xp = np.linspace(1, EMIR_NAXIS1, numpts)
    slit_dist = (slit_height * 10) + slit_gap

    # undistorted (constant) y-coordinate of the lower and upper boundaries
    ybottom = y_baseline * 100 + (islitlet - 1) * slit_dist
    ytop = ybottom + (slit_height * 10)

    list_spectrails = []
    for borderval in borderlist:
        yvalue = ybottom + borderval * (ytop - ybottom)
        # undistorted boundary
        yp_value = np.ones(numpts) * yvalue
        # distorted boundary
        xdist, ydist = exvp(xp, yp_value, x0=x0, y0=y0,
                            c2=c2, c4=c4, theta0=theta0, ff=ff)
        spectrail = SpectrumTrail()  # declare SpectrumTrail instance
        spectrail.fit(x=xdist, y=ydist, deg=deg, debugplot=debugplot)
        list_spectrails.append(spectrail)

    return list_spectrails


def expected_distorted_frontiers(islitlet, csu_bar_slit_center,
                                 params, parmodel,
                                 numpts, deg, debugplot=0):
    """Return expected frontiers as a list with two SpectrumTrail instances.

    Note that the frontiers are computed as the polynomials that extend
    the slitlet region defined by its boundaries, encompassing half the
    slit gap between consecutive slitlets.

    Parameters
    ----------
    islitlet : int
        Number of slitlet.
    csu_bar_slit_center : float
        CSU bar slit center, in mm.
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    numpts : int
        Number of points in which the X-range interval is subdivided
        before fitting the returned polynomials.
    deg : int
        Degree of the fitted polynomial.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    list_frontiers : list of SpectrumTrail objects
        List containing the two SpectrumTrail objects defining the
        slitlet frontiers (lower and upper, respectively).

    """

    c2, c4, ff, slit_gap, slit_height, theta0, x0, y0, y_baseline = \
        return_params(islitlet, csu_bar_slit_center, params, parmodel)

    xp = np.linspace(1, EMIR_NAXIS1, numpts)
    slit_dist = (slit_height * 10) + slit_gap

    # undistorted (constant) y-coordinate of the lower and upper frontiers
    ybottom = y_baseline * 100 + (islitlet - 1) * slit_dist - slit_gap / 2
    ytop = ybottom + (slit_height * 10) + slit_gap

    list_frontiers = []
    for i in range(2):
        if i == 0:
            yvalue = ybottom
        else:
            yvalue = ytop
        # undistorted boundary
        yp_value = np.ones(numpts) * yvalue
        # distorted boundary
        xdist, ydist = exvp(xp, yp_value, x0=x0, y0=y0,
                            c2=c2, c4=c4, theta0=theta0, ff=ff)
        spectrail = SpectrumTrail()  # declare SpectrumTrail instance
        spectrail.fit(x=xdist, y=ydist, deg=deg, debugplot=debugplot)
        list_frontiers.append(spectrail)

    return list_frontiers


def fun_residuals(params, parmodel, bounddict,
                  shrinking_factor, numresolution,
                  islitmin, islitmax, debugplot):
    """Function to be minimised.

    Parameters
    ----------
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    bounddict : JSON structure
        Structure employed to store bounddict information.
    shrinking_factor : float
        Fraction of the detected X range (specrtral) to be employed
        in the fit. This must be a number verifying
        0 < shrinking_factor <= 1. The resulting interval will be
        centered within the original one.
    numresolution : int
        Number of points in which the X-range interval is subdivided
        before computing the residuals.
    islitmin : int
        Minimum slitlet number.
    islitmax : int
        Maximum slitlet number.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    global_residual : float
        Squared root of the averaged sum of squared residuals.

    """

    global FUNCTION_EVALUATIONS
    global_residual = 0.0
    nsummed = 0

    read_slitlets = list(bounddict['contents'].keys())
    # read_slitlets.sort()  # this is not really necessary
    for tmp_slitlet in read_slitlets:
        islitlet = int(tmp_slitlet[7:])
        if islitmin <= islitlet <= islitmax:
            read_dateobs = list(bounddict['contents'][tmp_slitlet].keys())
            # read_dateobs.sort()  # this is not really necessary
            for tmp_dateobs in read_dateobs:
                tmp_dict = bounddict['contents'][tmp_slitlet][tmp_dateobs]
                csu_bar_slit_center = tmp_dict['csu_bar_slit_center']
                # expected boundaries using provided parameters
                list_spectrails = expected_distorted_boundaries(
                        islitlet, csu_bar_slit_center,
                        [0, 1], params, parmodel,
                        numpts=numresolution, deg=5, debugplot=0
                    )
                poly_lower_expected = list_spectrails[0].poly_funct
                poly_upper_expected = list_spectrails[1].poly_funct
                # measured lower boundary
                poly_lower_measured = np.polynomial.Polynomial(
                    tmp_dict['boundary_coef_lower']
                )
                xmin_lower_bound = tmp_dict['boundary_xmin_lower']
                xmax_lower_bound = tmp_dict['boundary_xmax_lower']
                dx = (xmax_lower_bound - xmin_lower_bound) * \
                     (1 - shrinking_factor) / 2
                xdum_lower = np.linspace(xmin_lower_bound + dx,
                                         xmax_lower_bound - dx,
                                         num=numresolution)
                # distance between expected and measured polynomials
                poly_diff = poly_lower_expected - poly_lower_measured
                global_residual += np.sum(poly_diff(xdum_lower)**2)
                nsummed += numresolution
                # measured upper boundary
                poly_upper_measured = np.polynomial.Polynomial(
                    tmp_dict['boundary_coef_upper']
                )
                xmin_upper_bound = tmp_dict['boundary_xmin_upper']
                xmax_upper_bound = tmp_dict['boundary_xmax_upper']
                dx = (xmax_lower_bound - xmin_lower_bound) * \
                     (1 - shrinking_factor) / 2
                xdum_upper = np.linspace(xmin_upper_bound + dx,
                                         xmax_upper_bound - dx,
                                         num=numresolution)
                # distance between expected and measured polynomials
                poly_diff = poly_upper_expected - poly_upper_measured
                global_residual += np.sum(poly_diff(xdum_upper)**2)
                nsummed += numresolution

    if nsummed > 0:
        global_residual = np.sqrt(global_residual/nsummed)
    if debugplot >= 10:
        FUNCTION_EVALUATIONS += 1
        print('-' * 79)
        print('>>> Number of function evaluations:', FUNCTION_EVALUATIONS)
        print('>>> global residual...............:', global_residual)
        params.pretty_print()
    return global_residual


def overplot_boundaries_from_bounddict(ax, bounddict, micolors, linetype='-'):
    """Overplot boundaries on current plot.

    Parameters
    ----------
    ax : matplotlib axes
        Current plot axes.
    bounddict : JSON structure
        Structure employed to store bounddict information.
    micolors : list of char
        List with two characters corresponding to alternating colors
        for odd and even slitlets.
    linetype : str
        Line type.

    """

    for islitlet in range(1, EMIR_NBARS + 1):
        tmpcolor = micolors[islitlet % 2]
        tmp_slitlet = 'slitlet' + str(islitlet).zfill(2)
        if tmp_slitlet in bounddict['contents'].keys():
            read_dateobs = list(bounddict['contents'][tmp_slitlet].keys())
            read_dateobs.sort()
            for tmp_dateobs in read_dateobs:
                tmp_dict = bounddict['contents'][tmp_slitlet][tmp_dateobs]
                # lower boundary
                pol_lower_measured = np.polynomial.Polynomial(
                    tmp_dict['boundary_coef_lower']
                )
                xdum = np.linspace(1, EMIR_NAXIS1, num=EMIR_NAXIS1)
                ydum = pol_lower_measured(xdum)
                ax.plot(xdum, ydum, tmpcolor + linetype)
                pol_upper_measured = np.polynomial.Polynomial(
                    tmp_dict['boundary_coef_upper']
                )
                ydum = pol_upper_measured(xdum)
                ax.plot(xdum, ydum, tmpcolor + linetype)


def overplot_boundaries_from_params(ax, params, parmodel,
                                    list_islitlet,
                                    list_csu_bar_slit_center,
                                    micolors=('m', 'c'), linetype='--',
                                    labels=True, alpha_fill=None,
                                    global_offset_x_pix=0,
                                    global_offset_y_pix=0):
    """Overplot boundaries computed from fitted parameters.

    Parameters
    ----------
    ax : matplotlib axes or None
        Current plot axes.
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    list_islitlet : list of integers
        Slitlet numbers to be considered.
        longslits.
    list_csu_bar_slit_center : list of floats
        CSU bar slit centers of the considered slitlets.
    micolors : Python list
        List with two characters corresponding to alternating colors
        for odd and even slitlets.
    linetype : str
        Line type.
    labels : bool
        If True, display slilet label
    alpha_fill : float or None
        Alpha factor to be employed to fill slitlet region.
    global_integer_offset_x_pix : int or float
        Global offset in the X direction to be applied after computing
        the expected location.
    global_offset_y_pix : int or float
        Global offset in the Y direction to be applied after computing
        the expected location.

    Returns
    -------
    list_pol_lower_boundaries : python list
        List of numpy.polynomial.Polynomial instances with the lower
        polynomial boundaries computed for the requested slitlets.
    list_pol_upper_boundaries : python list
        List of numpy.polynomial.Polynomial instances with the upper
        polynomial boundaries computed for the requested slitlets.


    """

    # duplicate to shorten the variable names
    xoff = float(global_offset_x_pix)
    yoff = float(global_offset_y_pix)

    list_pol_lower_boundaries = []
    list_pol_upper_boundaries = []
    for islitlet, csu_bar_slit_center in \
            zip(list_islitlet, list_csu_bar_slit_center):
        tmpcolor = micolors[islitlet % 2]
        pol_lower_expected = expected_distorted_boundaries(
            islitlet, csu_bar_slit_center,
            [0], params, parmodel, numpts=101, deg=5, debugplot=0
        )[0].poly_funct
        list_pol_lower_boundaries.append(pol_lower_expected)
        pol_upper_expected = expected_distorted_boundaries(
            islitlet, csu_bar_slit_center,
            [1], params, parmodel, numpts=101, deg=5, debugplot=0
        )[0].poly_funct
        list_pol_upper_boundaries.append(pol_upper_expected)

        if ax is not None:
            xdum = np.linspace(1, EMIR_NAXIS1, num=EMIR_NAXIS1)
            ydum1 = pol_lower_expected(xdum)
            ax.plot(xdum + xoff, ydum1 + yoff, tmpcolor + linetype)
            ydum2 = pol_upper_expected(xdum)
            ax.plot(xdum + xoff, ydum2 + yoff, tmpcolor + linetype)
            if alpha_fill is not None:
                ax.fill_between(xdum + xoff, ydum1 + yoff, ydum2 + yoff,
                                facecolor=tmpcolor, alpha=alpha_fill)
            if labels:
                # slitlet label
                yc_lower = pol_lower_expected(EMIR_NAXIS1 / 2 + 0.5)
                yc_upper = pol_upper_expected(EMIR_NAXIS1 / 2 + 0.5)
                xcsu = EMIR_NAXIS1 * csu_bar_slit_center / 341.5
                ax.text(xcsu + xoff, (yc_lower + yc_upper) / 2 + yoff,
                        str(islitlet),
                        fontsize=10, va='center', ha='center',
                        bbox=dict(boxstyle="round,pad=0.1", fc="white",
                                  ec="grey"),
                        color=tmpcolor, fontweight='bold',
                        backgroundcolor='white')

    # return lists with boundaries
    return list_pol_lower_boundaries, list_pol_upper_boundaries


def overplot_frontiers_from_params(ax, params, parmodel,
                                   list_islitlet,
                                   list_csu_bar_slit_center,
                                   micolors=('m', 'c'), linetype='--',
                                   labels=True, alpha_fill=None,
                                   global_offset_x_pix=0,
                                   global_offset_y_pix=0):
    """Overplot frontiers computed from fitted parameters.

    Parameters
    ----------
    ax : matplotlib axes or None
        Current plot axes.
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    list_islitlet : list of integers
        Slitlet numbers to be considered.
        longslits.
    list_csu_bar_slit_center : list of floats
        CSU bar slit centers of the considered slitlets.
    micolors : Python list
        List with two characters corresponding to alternating colors
        for odd and even slitlets.
    linetype : str
        Line type.
    labels : bool
        If True, display slilet label
    alpha_fill : float or None
        Alpha factor to be employed to fill slitlet region.
    global_integer_offset_x_pix : int or float
        Global offset in the X direction to be applied after computing
        the expected location.
    global_offset_y_pix : int or float
        Global offset in the Y direction to be applied after computing
        the expected location.

    Returns
    -------
    list_pol_lower_frontiers : python list
        List of SpectrumTrail instances with the lower polynomial
        frontiers computed for the requested slitlets.
    list_pol_upper_frontiers : python list
        List of SpectrumTrail instances with the upper polynomial
        frontiers computed for the requested slitlets.


    """

    # duplicate to shorten the variable names
    xoff = float(global_offset_x_pix)
    yoff = float(global_offset_y_pix)

    list_pol_lower_frontiers = []
    list_pol_upper_frontiers = []
    for islitlet, csu_bar_slit_center in \
            zip(list_islitlet, list_csu_bar_slit_center):
        tmpcolor = micolors[islitlet % 2]
        list_expected_frontiers = expected_distorted_frontiers(
            islitlet, csu_bar_slit_center,
            params, parmodel, numpts=101, deg=5, debugplot=0
        )
        pol_lower_expected = list_expected_frontiers[0].poly_funct
        list_pol_lower_frontiers.append(pol_lower_expected)
        pol_upper_expected = list_expected_frontiers[1].poly_funct
        list_pol_upper_frontiers.append(pol_upper_expected)
        if ax is not None:
            xdum = np.linspace(1, EMIR_NAXIS1, num=EMIR_NAXIS1)
            ydum1 = pol_lower_expected(xdum)
            ax.plot(xdum + xoff, ydum1 + yoff, tmpcolor + linetype)
            ydum2 = pol_upper_expected(xdum)
            ax.plot(xdum + xoff, ydum2 + yoff, tmpcolor + linetype)
            if alpha_fill is not None:
                ax.fill_between(xdum + xoff, ydum1 + yoff, ydum2 + yoff,
                                facecolor=tmpcolor, alpha=alpha_fill)
            if labels:
                # slitlet label
                yc_lower = pol_lower_expected(EMIR_NAXIS1 / 2 + 0.5)
                yc_upper = pol_upper_expected(EMIR_NAXIS1 / 2 + 0.5)
                xcsu = EMIR_NAXIS1 * csu_bar_slit_center / 341.5
                ax.text(xcsu + xoff, (yc_lower + yc_upper) / 2 + yoff,
                        str(islitlet),
                        fontsize=10, va='center', ha='center',
                        bbox=dict(boxstyle="round,pad=0.1", fc="white",
                                  ec="grey"),
                        color=tmpcolor, fontweight='bold',
                        backgroundcolor='white')

    # return lists of SpectrumTrail boundaries
    return list_pol_lower_frontiers, list_pol_upper_frontiers


def save_boundaries_from_bounddict_ds9(bounddict, ds9_filename, numpix=100):
    """Export to ds9 region file the boundaries in bounddict.

    Parameters
    ----------
    bounddict : JSON structure
        Structure employed to store bounddict information.
    ds9_filename : str
        Output file name for the ds9 region file.
    numpix : int
        Number of points in which the X-range interval is subdivided
        in order to save each boundary as a connected set of line
        segments.

    """

    ds9_file = open(ds9_filename, 'w')

    ds9_file.write('# Region file format: DS9 version 4.1\n')
    ds9_file.write('global color=green dashlist=2 4 width=2 '
                   'font="helvetica 10 normal roman" select=1 '
                   'highlite=1 dash=1 fixed=0 edit=1 '
                   'move=1 delete=1 include=1 source=1\n')
    ds9_file.write('physical\n')

    uuid = bounddict['uuid']
    spfilter = bounddict['tags']['filter']
    grism = bounddict['tags']['grism']

    ds9_file.write('#\n# uuid (boundict file): {0}\n'.format(uuid))
    ds9_file.write('# filter..............: {0}\n'.format(spfilter))
    ds9_file.write('# grism...............: {0}\n'.format(grism))

    colorbox = ['green', 'green']
    for islitlet in range(1, EMIR_NBARS + 1):
        tmp_slitlet = 'slitlet' + str(islitlet).zfill(2)
        if tmp_slitlet in bounddict['contents'].keys():
            ds9_file.write('#\n# islitlet: {0}\n'.format(tmp_slitlet))
            read_dateobs = list(bounddict['contents'][tmp_slitlet].keys())
            read_dateobs.sort()
            for tmp_dateobs in read_dateobs:
                ds9_file.write('#\n# date-obs: {0}\n'.format(tmp_dateobs))
                tmp_dict = bounddict['contents'][tmp_slitlet][tmp_dateobs]
                # lower boundary
                pol_lower_measured = np.polynomial.Polynomial(
                    tmp_dict['boundary_coef_lower']
                )
                xmin_lower = tmp_dict['boundary_xmin_lower']
                xmax_lower = tmp_dict['boundary_xmax_lower']
                xdum = np.linspace(xmin_lower, xmax_lower, num=numpix)
                ydum = pol_lower_measured(xdum)
                for i in range(len(xdum) - 1):
                    ds9_file.write(
                        'line {0} {1} {2} {3}'.format(xdum[i], ydum[i],
                                                      xdum[i + 1], ydum[i + 1])
                    )
                    ds9_file.write(
                        ' # color={0}\n'.format(colorbox[islitlet % 2]))
                # upper boundary
                pol_upper_measured = np.polynomial.Polynomial(
                    tmp_dict['boundary_coef_upper']
                )
                xmin_upper = tmp_dict['boundary_xmin_upper']
                xmax_upper = tmp_dict['boundary_xmax_upper']
                xdum = np.linspace(xmin_upper, xmax_upper, num=numpix)
                ydum = pol_upper_measured(xdum)
                for i in range(len(xdum) - 1):
                    ds9_file.write(
                        'line {0} {1} {2} {3}'.format(xdum[i], ydum[i],
                                                      xdum[i + 1], ydum[i + 1])
                    )
                    ds9_file.write(
                        ' # color={0}\n'.format(colorbox[islitlet % 2]))
                # slitlet label
                xlabel = xmax_lower + xmax_upper + xmin_lower + xmin_upper
                xlabel /= 4
                yc_lower = pol_lower_measured(xlabel)
                yc_upper = pol_upper_measured(xlabel)
                ds9_file.write('text {0} {1} {{{2}}} # color={3} '
                               'font="helvetica 10 bold '
                               'roman"\n'.format(xlabel,
                                                 (yc_lower + yc_upper) / 2,
                                                 islitlet,
                                                 colorbox[islitlet % 2]))

    ds9_file.close()


def save_boundaries_from_params_ds9(params, parmodel,
                                    list_islitlet,
                                    list_csu_bar_slit_center,
                                    uuid, grism, spfilter,
                                    ds9_filename, numpix=100,
                                    global_offset_x_pix=0,
                                    global_offset_y_pix=0):
    """Export to ds9 region file the boundaries parametrised with params.

    Parameters
    ----------
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    list_islitlet : list (integers)
        Slitlet numbers to be considered.
    list_csu_bar_slit_center : list of floats
        CSU bar slit centers of the considered slitlets.
    uuid: str
        UUID associated to the parameters 'params'.
    grism : str
        Employed grism.
    spfilter : str
        Employed filter.
    ds9_filename : str
        Output file name for the ds9 region file.
    numpix : int
        Number of points in which the X-range interval is subdivided
        in order to save each boundary as a connected set of line
        segments.
    global_offset_x_pix : int or float
        Global offset in the X direction to be applied after computing
        the expected location.
    global_offset_y_pix : int or float
        Global offset in the Y direction to be applied after computing
        the expected location.

    """

    ds9_file = open(ds9_filename, 'w')

    ds9_file.write('# Region file format: DS9 version 4.1\n')
    ds9_file.write('global color=green dashlist=2 4 width=2 '
                   'font="helvetica 10 normal roman" select=1 '
                   'highlite=1 dash=1 fixed=0 edit=1 '
                   'move=1 delete=1 include=1 source=1\n')
    ds9_file.write('physical\n#\n')

    ds9_file.write('#\n# uuid..: {0}\n'.format(uuid))
    ds9_file.write('# filter: {0}\n'.format(spfilter))
    ds9_file.write('# grism.: {0}\n'.format(grism))

    ds9_file.write('#\n# global_offset_x_pix: {0}\n'.format(
        global_offset_x_pix))
    ds9_file.write('# global_offset_y_pix: {0}\n#\n'.format(
        global_offset_y_pix))

    # duplicate to shorten the variable names
    xoff = float(global_offset_x_pix)
    yoff = float(global_offset_y_pix)

    if parmodel == "longslit":
        for dumpar in EXPECTED_PARAMETER_LIST:
            parvalue = params[dumpar].value
            ds9_file.write('# {0}: {1}\n'.format(dumpar, parvalue))
    else:
        for dumpar in EXPECTED_PARAMETER_LIST_EXTENDED:
            parvalue = params[dumpar].value
            ds9_file.write('# {0}: {1}\n'.format(dumpar, parvalue))

    for islitlet, csu_bar_slit_center in \
            zip(list_islitlet, list_csu_bar_slit_center):
        if islitlet % 2 == 0:
            colorbox = '#ff00ff'  # '#ff77ff'
        else:
            colorbox = '#00ffff'  # '#4444ff'

        ds9_file.write(
            '#\n# islitlet...........: {0}\n'.format(islitlet)
        )
        ds9_file.write(
            '# csu_bar_slit_center: {0}\n'.format(csu_bar_slit_center)
        )
        pol_lower_expected = expected_distorted_boundaries(
            islitlet, csu_bar_slit_center, [0], params, parmodel,
            numpts=101, deg=5, debugplot=0
        )[0].poly_funct
        pol_upper_expected = expected_distorted_boundaries(
            islitlet, csu_bar_slit_center, [1], params, parmodel,
            numpts=101, deg=5, debugplot=0
        )[0].poly_funct
        xdum = np.linspace(1, EMIR_NAXIS1, num=numpix)
        ydum = pol_lower_expected(xdum)
        for i in range(len(xdum)-1):
            ds9_file.write(
                'line {0} {1} {2} {3}'.format(xdum[i]+xoff, ydum[i]+yoff,
                                              xdum[i+1]+xoff, ydum[i+1]+yoff)
            )
            ds9_file.write(' # color={0}\n'.format(colorbox))
        ydum = pol_upper_expected(xdum)
        for i in range(len(xdum)-1):
            ds9_file.write(
                'line {0} {1} {2} {3}'.format(xdum[i]+xoff, ydum[i]+yoff,
                                              xdum[i+1]+xoff, ydum[i+1]+yoff)
            )
            ds9_file.write(' # color={0}\n'.format(colorbox))
        # slitlet label
        yc_lower = pol_lower_expected(EMIR_NAXIS1 / 2 + 0.5)
        yc_upper = pol_upper_expected(EMIR_NAXIS1 / 2 + 0.5)
        ds9_file.write('text {0} {1} {{{2}}} # color={3} '
                       'font="helvetica 10 bold '
                       'roman"\n'.format(EMIR_NAXIS1 / 2 + 0.5 + xoff,
                                         (yc_lower + yc_upper) / 2 + yoff,
                                         islitlet,
                                         colorbox))

    ds9_file.close()


def save_frontiers_from_params_ds9(params, parmodel,
                                   list_islitlet,
                                   list_csu_bar_slit_center,
                                   uuid, grism, spfilter,
                                   ds9_filename, numpix=100,
                                   global_offset_x_pix=0,
                                   global_offset_y_pix=0):
    """Export to ds9 region file the frontiers parametrised with params.

    Parameters
    ----------
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    list_islitlet : list (integers)
        Slitlet numbers to be considered.
    list_csu_bar_slit_center : list of floats
        CSU bar slit centers of the considered slitlets.
    uuid: str
        UUID associated to the parameters 'params'.
    grism : str
        Employed grism.
    spfilter : str
        Employed filter.
    ds9_filename : str
        Output file name for the ds9 region file.
    numpix : int
        Number of points in which the X-range interval is subdivided
        in order to save each boundary as a connected set of line
        segments.
    global_integer_offset_x_pix : int or float
        Global offset in the X direction to be applied after computing
        the expected location.
    global_offset_y_pix : int or float
        Global offset in the Y direction to be applied after computing
        the expected location.

    """

    ds9_file = open(ds9_filename, 'w')

    ds9_file.write('# Region file format: DS9 version 4.1\n')
    ds9_file.write('global color=green dashlist=2 4 width=2 '
                   'font="helvetica 10 normal roman" select=1 '
                   'highlite=1 dash=1 fixed=0 edit=1 '
                   'move=1 delete=1 include=1 source=1\n')
    ds9_file.write('physical\n#\n')

    ds9_file.write('#\n# uuid..: {0}\n'.format(uuid))
    ds9_file.write('# filter: {0}\n'.format(spfilter))
    ds9_file.write('# grism.: {0}\n'.format(grism))

    ds9_file.write('#\n# global_offset_x_pix: {0}\n'.format(
        global_offset_x_pix))
    ds9_file.write('# global_offset_y_pix: {0}\n#\n'.format(
        global_offset_y_pix))

    # duplicate to shorten the variable names
    xoff = float(global_offset_x_pix)
    yoff = float(global_offset_y_pix)

    if parmodel == "longslit":
        for dumpar in EXPECTED_PARAMETER_LIST:
            parvalue = params[dumpar].value
            ds9_file.write('# {0}: {1}\n'.format(dumpar, parvalue))
    else:
        for dumpar in EXPECTED_PARAMETER_LIST_EXTENDED:
            parvalue = params[dumpar].value
            ds9_file.write('# {0}: {1}\n'.format(dumpar, parvalue))

    for islitlet, csu_bar_slit_center in \
            zip(list_islitlet, list_csu_bar_slit_center):
        if islitlet % 2 == 0:
            colorbox = '#0000ff'  # '#ff77ff'
        else:
            colorbox = '#0000ff'  # '#4444ff'

        ds9_file.write(
            '#\n# islitlet...........: {0}\n'.format(islitlet)
        )
        ds9_file.write(
            '# csu_bar_slit_center: {0}\n'.format(csu_bar_slit_center)
        )
        list_expected_frontiers = expected_distorted_frontiers(
            islitlet, csu_bar_slit_center,
            params, parmodel, numpts=101, deg=5, debugplot=0
        )
        pol_lower_expected = list_expected_frontiers[0].poly_funct
        pol_upper_expected = list_expected_frontiers[1].poly_funct
        xdum = np.linspace(1, EMIR_NAXIS1, num=numpix)
        ydum = pol_lower_expected(xdum)
        for i in range(len(xdum)-1):
            ds9_file.write(
                'line {0} {1} {2} {3}'.format(xdum[i]+xoff, ydum[i]+yoff,
                                              xdum[i+1]+xoff, ydum[i+1]+yoff)
            )
            ds9_file.write(' # color={0}\n'.format(colorbox))
        ydum = pol_upper_expected(xdum)
        for i in range(len(xdum)-1):
            ds9_file.write(
                'line {0} {1} {2} {3}'.format(xdum[i]+xoff, ydum[i]+yoff,
                                              xdum[i+1]+xoff, ydum[i+1]+yoff)
            )
            ds9_file.write(' # color={0}\n'.format(colorbox))
        # slitlet label
        yc_lower = pol_lower_expected(EMIR_NAXIS1 / 2 + 0.5)
        yc_upper = pol_upper_expected(EMIR_NAXIS1 / 2 + 0.5)
        ds9_file.write('text {0} {1} {{{2}}} # color={3} '
                       'font="helvetica 10 bold '
                       'roman"\n'.format(EMIR_NAXIS1 / 2 + 0.5 + xoff,
                                         (yc_lower + yc_upper) / 2 + yoff,
                                         islitlet,
                                         colorbox))

    ds9_file.close()


def bound_params_from_dict(bound_param_dict):
    """Define `~lmfit.parameter.Parameters` object from dictionary.

    Parameters
    ----------
    bound_param_dict : dictionary
        Dictionary containing the JSON contents of a boundary
        parameter file.

    Returns
    -------
    params : :class:`~lmfit.parameter.Parameters`
        Parameters object.

    """

    params = Parameters()
    for mainpar in EXPECTED_PARAMETER_LIST:
        if mainpar not in bound_param_dict['contents'].keys():
            raise ValueError('Parameter ' + mainpar + ' not found!')
        if bound_param_dict['meta_info']['parmodel'] == "longslit":
            dumdict = bound_param_dict['contents'][mainpar]
            params.add(mainpar, value=dumdict["value"],
                       vary=dumdict["vary"])
        elif bound_param_dict['meta_info']['parmodel'] == 'multislit':
            for subpar in ['a0s', 'a1s', 'a2s']:
                if subpar not in bound_param_dict['contents'][mainpar].keys():
                    raise ValueError('Subparameter ' + subpar + ' not found' +
                                     ' under parameter ' + mainpar)
                cpar = mainpar + '_' + subpar
                dumdict = bound_param_dict['contents'][mainpar][subpar]
                params.add(cpar, value=dumdict["value"],
                           vary=dumdict["vary"])
        else:
            print('parmodel: ', bound_param_dict['meta_info']['parmodel'])
            raise ValueError('Unexpected parmodel')
    return params


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()

    # positional arguments
    parser.add_argument("bounddict",
                        help="Input JSON boundary file with fits to "
                             "continuum-lamp exposures",
                        type=argparse.FileType('rt'))
    parser.add_argument("--init_bound_param", required=True,
                        help="Input JSON with initial boundary parameters",
                        type=argparse.FileType('rt'))
    parser.add_argument("--parmodel", required=True,
                        help="Parameter model: multislit (default) or "
                             "longslit",
                        default="multislit",
                        choices=("multislit", "longslit"))
    parser.add_argument("--fitted_bound_param", required=True,
                        help="Output JSON with fitted boundary parameters",
                        type=lambda x: arg_file_is_new(parser, x))

    # optional arguments
    parser.add_argument("--shrinking_factor",
                        help="Effective reduction factor to be applied to "
                             "the fitted X-axis (spectral) range, "
                             "(default=0.9)",
                        type=float, default=0.9)
    parser.add_argument("--tolerance",
                        help="Tolerance for Nelder-Mead minimization process "
                             "(default=1E-7)",
                        type=float, default=1e-7)
    parser.add_argument("--maxDTUoffset",
                        help="Maximum allowed difference in DTU location (mm)"
                             "for each parameter (default=0.5)",
                        type=float, default=0.5)
    parser.add_argument("--numresolution",
                        help="Number of points/boundary (default=101)",
                        type=int, default=101)
    parser.add_argument("--background_image",
                        help="Optional FITS image to display as background "
                             "image",
                        type=argparse.FileType('rb'))
    parser.add_argument("--pickle_input",
                        help="Use previous pickle file instead of carrying "
                             "out the minimisation procces",
                        type=argparse.FileType('rb'))
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                             " (default=0)",
                        type=int, default=0,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    if args.background_image is not None and args.debugplot % 11 == 0:
        raise ValueError("--background_image requires "
                         "--debugplot value compatible with "
                         "plotting\n'")

    if args.shrinking_factor <= 0 or args.shrinking_factor > 1:
        raise ValueError("Unexpected shriking factor: ",
                         args.shrinking_factor, '\n')

    # read bounddict file and check its contents
    bounddict = json.loads(open(args.bounddict.name).read())
    averaged_dtu_configuration, maxdiff_dtu_configuration = \
        integrity_check(bounddict, args.maxDTUoffset)
    save_boundaries_from_bounddict_ds9(bounddict, 'ds9_bounddict.reg')
    # store lists with individual slitlet number and csu_bar_slit_center
    # value, needed later to save the ds9 region file and for plotting
    list_islitlet = []
    list_csu_bar_slit_center = []
    read_slitlets = list(bounddict['contents'].keys())
    read_slitlets.sort()
    for tmp_slitlet in read_slitlets:
        islitlet = int(tmp_slitlet[7:])
        list_islitlet.append(islitlet)
        read_dateobs = list(bounddict['contents'][tmp_slitlet].keys())
        read_dateobs.sort()
        for tmp_dateobs in read_dateobs:
            tmp_dict = bounddict['contents'][tmp_slitlet][tmp_dateobs]
            csu_bar_slit_center = tmp_dict['csu_bar_slit_center']
            list_csu_bar_slit_center.append(csu_bar_slit_center)

    uuid = bounddict['uuid']
    grism = bounddict['tags']['grism']
    spfilter = bounddict['tags']['filter']

    # read init_bound_param file
    init_bound_param = json.loads(open(args.init_bound_param.name).read())
    # check that grism and filter match
    grism_ = init_bound_param['tags']['grism']
    spfilter_ = init_bound_param['tags']['filter']
    if grism != grism_:
        print("grism (JSON bounddict..):", grism)
        print("grism (init_bound_param):", grism_)
        raise ValueError("grism mismatch")
    if spfilter != spfilter_:
        print("filter (JSON bounddict..):", spfilter)
        print("filter (init_bound_param):", spfilter_)
        raise ValueError("filter mismatch")
    islitlet_min = init_bound_param['tags']['islitlet_min']
    islitlet_max = init_bound_param['tags']['islitlet_max']
    # check that parameter model is correct
    parmodel_ = init_bound_param['meta_info']['parmodel']
    if args.parmodel != parmodel_:
        raise ValueError("Unexpected parmodel: ", parmodel_,
                         " in file ", args.init_bound_param.name)

    # establish initial parameters from init_bound_param dictionary
    params = bound_params_from_dict(init_bound_param)

    print('-' * 79)
    print('* INITIAL PARAMETERS')
    params.pretty_print()
    pause_debugplot(args.debugplot)

    # carry out minimisation process (or read previous pickle file)
    if args.pickle_input is not None:
        result = pickle.load(args.pickle_input)
    else:
        fitter = Minimizer(
            fun_residuals, params,
            fcn_args=(args.parmodel, bounddict,
                      args.shrinking_factor, args.numresolution,
                      islitlet_min, islitlet_max, args.debugplot)
        )
        result = fitter.scalar_minimize(method='Nelder-Mead',
                                        tol=args.tolerance)
        pickle.dump(result, open('dum.pickle', 'wb'))

    global_residual = fun_residuals(result.params, args.parmodel, bounddict,
                                    args.shrinking_factor,
                                    args.numresolution,
                                    islitlet_min, islitlet_max,
                                    args.debugplot)
    print('\n>>> global residual', global_residual)
    result.params.pretty_print()
    print('>>> Number of function evaluations:', result.nfev)
    pause_debugplot(args.debugplot)
    report_fit(result.params, min_correl=0.5)
    pause_debugplot(args.debugplot)

    # export resulting boundaries to ds9 region file for the longslit
    # case (otherwise there is not a single csu_bar_slit_center)
    if args.parmodel == "longslit":
        save_boundaries_from_params_ds9(result.params, args.parmodel,
                                        list_islitlet,
                                        list_csu_bar_slit_center,
                                        uuid, grism, spfilter,
                                        'ds9_fittedpar.reg')

    fitted_bound_param = deepcopy(init_bound_param)
    fitted_bound_param['meta_info']['creation_date'] = \
        datetime.now().isoformat()
    fitted_bound_param['meta_info']['description'] \
        = "fitted boundary parameters"
    fitted_bound_param['meta_info']['function_evaluations'] = result.nfev
    fitted_bound_param['meta_info']['global_residual'] = global_residual
    fitted_bound_param['meta_info']['numresolution'] = args.numresolution
    fitted_bound_param['meta_info']['tolerance'] = args.tolerance
    fitted_bound_param['meta_info']['maxDTUoffset'] = args.maxDTUoffset
    fitted_bound_param['meta_info']['origin'] = {}
    fitted_bound_param['meta_info']['origin']['bounddict_uuid'] = \
        bounddict['uuid']
    fitted_bound_param['meta_info']['origin']['init_bound_param_uuid'] = \
        init_bound_param['uuid']
    fitted_bound_param['dtu_configuration'] = \
        averaged_dtu_configuration.outdict()
    fitted_bound_param['dtu_configuration_maxdiff'] = \
        maxdiff_dtu_configuration.outdict()
    fitted_bound_param['uuid'] = str(uuid4())
    for mainpar in EXPECTED_PARAMETER_LIST:
        if mainpar not in init_bound_param['contents'].keys():
            raise ValueError('Parameter ' + mainpar +
                             ' not found in ' + args.init_bound_param.name)
        if args.parmodel == "longslit":
            fitted_bound_param['contents'][mainpar]['value'] = \
                result.params[mainpar].value
            fitted_bound_param['contents'][mainpar]['initial'] = \
                params[mainpar].value
            # compute median csu_bar_slit_center (only in longslit mode)
            dumlist = []
            for islitlet, csu_bar_slit_center in \
                    zip(list_islitlet, list_csu_bar_slit_center):
                if islitlet_min <= islitlet <= islitlet_max:
                    dumlist.append(csu_bar_slit_center)
            median_csu_bar_slit_center = np.median(np.array(dumlist))
            fitted_bound_param['meta_info']['median_csu_bar_slit_center']\
                = median_csu_bar_slit_center
        else:
            for subpar in ['a0s', 'a1s', 'a2s']:
                if subpar not in \
                        init_bound_param['contents'][mainpar].keys():
                    raise ValueError(
                        'Subparameter ' + subpar + ' not found in ' +
                        args.init_bound_param.name + ' under parameter ' +
                        mainpar
                    )
                cpar = mainpar + '_' + subpar
                fitted_bound_param['contents'][mainpar][subpar]['value']\
                    = result.params[cpar].value
                fitted_bound_param['contents'][mainpar][subpar]['initial']\
                    = params[cpar].value
    with open(args.fitted_bound_param.name, 'w') as fstream:
        json.dump(fitted_bound_param, fstream, indent=2, sort_keys=True)

    if args.debugplot % 10 != 0:
        fig = plt.figure()
        geometry = (0, 0, 640, 480)
        set_window_geometry(geometry)
        if args.background_image is not None:
            # read input FITS file
            hdulist = fits.open(args.background_image.name)
            image2d = hdulist[0].data
            hdulist.close()
            if image2d.shape != (EMIR_NAXIS2, EMIR_NAXIS1):
                raise ValueError("Unexpected error with NAXIS1, NAXIS2")
            ax = ximshow(image2d=image2d,
                         title=args.background_image.name,
                         image_bbox=(1, EMIR_NAXIS1, 1, EMIR_NAXIS2),
                         show=False)
        else:
            ax = fig.add_subplot(111)
            ax.set_xlim(-0.5, EMIR_NAXIS1 + 0.5)
            ax.set_ylim(-0.5, EMIR_NAXIS2 + 0.5)
            ax.set_xlabel('X axis (from 1 to NAXIS1)')
            ax.set_xlabel('Y axis (from 1 to NAXIS2)')
            ax.set_title(args.bounddict.name)
        # boundaries from bounddict
        overplot_boundaries_from_bounddict(ax, bounddict, ['r', 'b'])
        # expected boundaries for the longslit case
        if args.parmodel == "longslit":
            overplot_boundaries_from_params(ax, result.params, args.parmodel,
                                            list_islitlet,
                                            list_csu_bar_slit_center)
        pause_debugplot(debugplot=args.debugplot, pltshow=True,
                        tight_layout=True)


if __name__ == "__main__":

    main()



#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Arc calibration Recipe for EMIR
"""

from __future__ import division, print_function

from numina.array.combine import median
from numina.core import Requirement, Result
from numina.types.linescatalog import LinesCatalog

from emirdrp.core.recipe import EmirRecipe
import emirdrp.decorators
from numina.processing.combine import basic_processing_with_combination
from emirdrp.processing.wavecal.apply_rectwv_coeff import apply_rectwv_coeff
from emirdrp.processing.wavecal.rectwv_coeff_from_arc_image \
    import rectwv_coeff_from_arc_image
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 import save_four_ds9
import emirdrp.products as prods
import emirdrp.requirements as reqs


class ArcCalibrationRecipe(EmirRecipe):
    """Process arc images applying wavelength calibration"""

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    bound_param = reqs.RefinedBoundaryModelParamRequirement()
    lines_catalog = Requirement(LinesCatalog, 'Catalog of lines')

    reduced_image = Result(prods.ProcessedImage)
    rectwv_coeff = Result(prods.RectWaveCoeff)
    reduced_55sp = Result(prods.ProcessedMOS)
    reduced_arc = Result(prods.ProcessedMOS)

    @emirdrp.decorators.loginfo
    def run(self, rinput):
        self.logger.info('starting rect.+wavecal. reduction of arc spectra')

        # build object to proceed with bpm, bias, dark and flat
        flow = self.init_filters(rinput)

        # apply bpm, bias, dark and flat
        reduced_image = basic_processing_with_combination(rinput, flow,
                                                          method=median)
        # update header con additional info
        hdr = reduced_image[0].header
        self.set_base_headers(hdr)

        # save intermediate image in work directory
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        # RectWaveCoeff object (with rectification and wavelength calibration
        # coefficients for the particular CSU configuration of the arc image)
        # and HDUList object with the FITS image corresponding to 55 median
        # spectra of each slitlet
        rectwv_coeff, reduced_55sp = rectwv_coeff_from_arc_image(
            reduced_image,
            rinput.bound_param,
            rinput.lines_catalog,
        )

        # generate associated ds9 region files and save them in work directory
        if self.intermediate_results:
            save_four_ds9(rectwv_coeff)

        # apply rectification and wavelength calibration
        reduced_arc = apply_rectwv_coeff(reduced_image, rectwv_coeff)

        # save results in result directory
        self.logger.info('end rect.+wavecal. reduction of arc spectra')
        result = self.create_result(reduced_image=reduced_image,
                                    rectwv_coeff=rectwv_coeff,
                                    reduced_55sp=reduced_55sp,
                                    reduced_arc=reduced_arc)
        return result

    def set_base_headers(self, hdr):
        newhdr = super(ArcCalibrationRecipe, self).set_base_headers(hdr)
        # Update SEC to 0
        newhdr['SEC'] = 0
        return newhdr
